import pandas as pd
from ...data.struct import AssetWeight, AssetPrice, AssetPosition, AssetValue
from ...data.struct import TAAParam, FAParam, FundWeightItem, FundWeight, TaaTunerParam
from . import Helper


class SAAHelper(Helper):

    def __init__(self):
        pass

    def setup(self, saa: AssetWeight):
        self.saa = saa

    def on_price(self, dt, asset_price: AssetPrice):
        cur_saa = self.saa.copy()
        for k, v in asset_price.__dict__.items():
            if asset_price.isnan(v):
                cur_saa.__dict__[k] = 0
        cur_saa.rebalance()
        return cur_saa


class TAAStatusMode:
    NORMAL = 'normal'
    IN_LOW = 'low'
    IN_HIGH = 'high'

class TAAHelper(Helper):

    # 1. let each asset use same taa param: set taa_params and taa_param_details == None
    # 2. let some asset use special taa param and others default: set taa_params and make special params in taa_param_details
    # 3. stop some asset from using taa: make taa_param_details[index_id] = None
    # 4. only asset in taa_param_details has taa: make taa_params = None
    def __init__(self, taa_params: TAAParam=None, taa_param_details: dict=None):
        self.params = taa_params or TAAParam()
        self.param_details = taa_param_details or {}
        self.tactic_status = {}
        self.tactic_history = {}

    def on_price(self, dt, asset_price: AssetPrice, cur_saa: AssetWeight, asset_pct: pd.DataFrame, select_val: dict, score_dict: dict):
        taa = cur_saa.copy()
        taa_effected = False
        mode_changed = False
        tactic_dt = {}
        for index_id, target_w in taa.__dict__.items():
            # 如果self.param_details存在， 按照里面设定index做taa, 不设定的不做taa
            # 如果self.param_details不存在，默认值是{}, 则所有index统一参数 self.params
            cur_params = self.params if not bool(self.param_details) else self.param_details.get(index_id, None)
            if target_w == 0 or cur_params is None:
                continue
            #if index_id in TaaTunerParam.POOL:
            #assert index_id in asset_pct.index, f'{index_id} pct not exsit'
            # 无估值跳过  大部分资产08年之前没有估值百分位
            if index_id not in asset_pct.index:
                continue
            assert (index_id in select_val) or (index_id in TaaTunerParam.POOL), \
                f'index_id {index_id} without definded index valuation to taa'
            val_col = select_val[index_id] if index_id in select_val.keys() else TaaTunerParam.POOL[index_id]
            val_pct = asset_pct.loc[index_id, val_col]
            cur_mode = self.tactic_status.get(index_id, TAAStatusMode.NORMAL)
            new_mode = TAAStatusMode.NORMAL
            tactic_w = target_w

            if cur_mode == TAAStatusMode.NORMAL:

                if val_pct >= cur_params.HighThreshold:
                    tactic_w = max(target_w - cur_params.HighMinus, 0)
                    new_mode = TAAStatusMode.IN_HIGH
                elif val_pct <= cur_params.LowThreshold:
                    tactic_w = min(target_w + cur_params.LowPlus, 1)
                    new_mode = TAAStatusMode.IN_LOW
                
            elif cur_mode == TAAStatusMode.IN_LOW:

                if val_pct < cur_params.LowStop:
                    tactic_w = min(target_w + cur_params.LowPlus, 1)
                    new_mode = TAAStatusMode.IN_LOW

            elif cur_mode == TAAStatusMode.IN_HIGH:

                if val_pct > cur_params.HighStop:
                    tactic_w = max(target_w - cur_params.HighMinus, 0)
                    new_mode = TAAStatusMode.IN_HIGH
            else:
                assert False, 'should not be here!'

            self.tactic_status[index_id] = new_mode
            if new_mode != TAAStatusMode.NORMAL:
                taa.__dict__[index_id] = tactic_w
                if cur_params.ToMmf:
                    taa.mmf -= (tactic_w - target_w)

            mode_changed = mode_changed or new_mode != cur_mode
            taa_effected = taa_effected or new_mode != TAAStatusMode.NORMAL
            tactic_dt[index_id] = new_mode
        self.tactic_history[dt] = tactic_dt
        if taa.mmf < 0:
            #print(f'{dt} taa mmf less than 0 mmf weight {taa.mmf} tactic_dt {tactic_dt}')
            taa.mmf = 0
        self.rebalance_taa(taa, self.params.TuneCash)
        self.asset_adjust(taa, score_dict)
        if taa_effected:
            #print(f'taa: {dt} (mode){mode_changed} (taa){taa_effected} (saa){cur_saa} (taa){taa}')
            pass
        return taa

    def rebalance_taa(self, asset_weight:AssetWeight, to_tune_cash:bool):
        asset_weight = asset_weight.__dict__
        _sum = 0.0
        for k, v in asset_weight.items():
            assert round(v,8) >= 0, f'{k} has negative value'
            _sum += v
        assert _sum > 0, 'param sum value must be positive'
        # 验证过的非现金类资产rebalance
        if not to_tune_cash:
            if asset_weight['cash'] != _sum:
                _sum = (_sum - asset_weight['cash']) / (1 - asset_weight['cash'])
            # 资产配置只有现金时，不需要rebalance
            else:
                return
        # 不调整cash
        for k, v in asset_weight.items():
            if k == 'cash' and (not to_tune_cash):
                continue
            asset_weight[k] /= _sum

    def asset_adjust(self, asset_weight:AssetWeight, score_dict:dict):
        # 如果当前资产没有候选基金，买货基
        asset_weight = asset_weight.__dict__
        for k,v in asset_weight.items():
            if k in score_dict:
                index_fund_nums = len(list(score_dict[k].keys()))
                if index_fund_nums == 0:
                    asset_weight['mmf'] += v
                    asset_weight[k] -= v
 
class FAHelper(Helper):
    
    def __init__(self, fa_params: FAParam=None):
        self.params = fa_params or FAParam()
        self.proper_fund_numbers = {}

    def setup(self, saa: AssetWeight):
        self.saa = saa
        ## set proper fund num for each 
        self.fund_nums = {}
        for index_id, asset_wgt in self.saa.__dict__.items():
            if index_id == 'active':
                self.fund_nums[index_id] = 10
                continue
            max_fund_num = int(round(asset_wgt / 0.05, 0))
            if asset_wgt > 0 and max_fund_num <= 0:
                max_fund_num = 1
            max_fund_num = min(max_fund_num, self.params.MaxFundNumUnderAsset)
            if max_fund_num > 0:
                self.fund_nums[index_id] = max_fund_num
            
    def get_max_fund_num(self, index_id):
        return int(self.fund_nums.get(index_id, 1))        

    def on_price(self, dt, cur_asset_allocation: AssetWeight, cur_fund_score: dict, cur_manager_score_cleaned : dict, score_select_dict : dict):
        
        # if no fund in index_id, rebalance again
        for index_id, w in cur_asset_allocation.__dict__.items():
            if w == 0 or index_id == 'cash':
                continue
            _score_dic = cur_fund_score.get(index_id,None)
            if not _score_dic:
                cur_asset_allocation.__dict__[index_id] = 0
            else:
                _score_dic = _score_dic[score_select_dict[index_id]]
                if not _score_dic:
                    cur_asset_allocation.__dict__[index_id] = 0
        cur_asset_allocation.rebalance()

        res = FundWeight()
        for index_id in cur_asset_allocation.__dict__.keys():
            if index_id == 'cash':
                continue
            score_type = score_select_dict[index_id]
            asset_wgt = cur_asset_allocation.__dict__[index_id]
            if asset_wgt > 0:
                # 保证主动基金不买到相同基金 的基金经理
                if index_id == 'active':
                    # print(f'cur_manager_score_cleaned {cur_manager_score_cleaned}')
                    # print(f'score_type {score_type}')
                    _scores = sorted(cur_manager_score_cleaned[score_type].items(), key=lambda item: item[1], reverse=True)
                else:     
                    _scores = sorted(cur_fund_score.get(index_id, {}).get(score_type, {}).items(), key=lambda item: item[1], reverse=True)
                proper_fund_num = self.get_max_fund_num(index_id) # for a single asset, for each 5% wgt it gots, they should take 1 extra fund

                if proper_fund_num > len(_scores):
                    #print(f'lack fund candidates: (dt){dt} (index_id){index_id} (proper_num){proper_fund_num} (fund_num){len(_scores)}')
                    proper_fund_num = len(_scores)
                for i in range(0, proper_fund_num):
                    fund_id, _score = _scores[i]
                    fund_weight_item = FundWeightItem(fund_id=fund_id, index_id=index_id, asset_wgt=asset_wgt, fund_wgt_in_asset=1.0 / proper_fund_num)
                    res.add(fund_weight_item)

                #assert proper_fund_num > 0, f'no fund is selected in {index_id}'                    
        return res


from .trader import FundTrader

from ...data.struct import AssetWeight, AssetPrice, AssetPosition, AssetValue
from ...data.struct import AssetTrade, FundTrade, FundPosition, FundWeight
from ...data.struct import AssetTradeParam, FundTradeParam
from ...data.struct import AssetTradeCache, FundWeightItem
from ...data.constant import TradeTrigger
from . import Helper
from .asset_helper import FAHelper
import numpy as np
import math


class FundScoreTolTrader(FundTrader):
    '''
    判定是否需要交易

    核心是计算 index 偏移（index_max_diff）和 fund 偏移（fund_max_diff）

    最终用 diff = lambda * fund_max_diff + (1-lambda) * index_max_diff 是否超过阈值来判定

    FundTradeParam.DiffJudgeLambda
    FundTradeParam.DiffJudgeTolarance
    FundTradeParam.DiffJudgeAssetWgtRequirement

    loop 每一个待交易的资产：(index_id)
    1. 计算该资产需要调整的比例，如果大于 index_max_diff ，则更新该值
    2. 计算该资产下应该有的基金数量 （FAHelper.get_proper_fund_num）
    3. 计算按照当前的基金排名情况，应该实际持有哪些基金（new_fund_ids）
    4. 按照 new_fund_ids 等权分布，统计该资产下所有基金需要变更的比例之和（暂时只统计买入单边，不考虑该资产占总比的情况）该资产占比超过 3%（DiffJudgeAssetWgtRequirement） 且该比例大于 fund_max_diff，则更新该值
    # 此时我们获得了 index_max_diff 和 fund_max_diff

    最终交易目标：
    对于某个资产，下面应该有 proper_num 个基金

    1. 当前持仓仍排名在前 proper_num 的基金，我们统一安排 1 / proper_num 的比例
    2. 如果排名在 proper_num + 1 ~ 2 * proper_num 的基金，如果不足 1 / proper_num 的比例，则不变，如果多于 1 / proper_num，就减至 1 / proper_num。（只减少，不增加）
    3. 如果当前已安排的基金少于 proper_num，则增加排名在前 proper_num 但不在当前列表中的 proper_num - current_num 个基金到列表中，每个安排 1 / proper_num 的比例
    4. 如果当前比例还有剩余（步骤2中产生的），如果步骤1中前 proper_num 有哪个减少过，就按照排名高低revert这个减少；如果将减少的都处理完还有剩余，那么进入到特殊状态（SPECIAL）
    5. 对于SPECIAL，我们先看fund_selected_num是否已达到proper_num，如果否，从fund_score中按评分从高到低将不在当前持仓的基金每个都增加进去，最多不超过 1 / proper_num，直到fund_selected_num等于proper_num或剩余的量消耗光；如果是，或fund_selected_num等于proper_num时剩余的量仍未消耗光，我们将剩余的比例平均分配给当前持仓的基金，以此来将剩余的量消耗光（分配后有可能超过 1 / proper_num，但实际在上边第4条中也有可能导致超过 1 / proper_num）
    '''

    def get_trades(self, dt,
            index_fund_cache: AssetTradeCache, 
            tar_fund_weight: FundWeight,
            tar_asset_weight: AssetWeight,
            cur_fund_position: FundPosition,
            cur_asset_position: AssetPosition,
            cur_fund_nav: dict,
            cur_fund_score: dict,
            cur_asset_price: AssetPrice):

        fund_tot_mv, cur_fund_wgts = cur_fund_position.calc_mv_n_w(fund_navs=cur_fund_nav)

        fund_trades = []

        for c in index_fund_cache:
            funds = {}
            fund_selected_num = 0
            fund_selected_wgt = 0
            proper_fund_wgt = 1.0 / c.proper_fund_num if c.proper_fund_num > 0 else 0
            good_fund_to_sell = []
            to_cancel = set([])
            # 如果 cur_fund_ids 超过 proper_fund_num
            if len(c.cur_fund_ids) > c.proper_fund_num:
                _d = {fund_id: c.fund_ranks.get(fund_id) for fund_id in c.cur_fund_ids}
                _dd = sorted(_d.items(), key=lambda x: x[1], reverse=True)
                for fund_id, _rank in _dd[:len(c.cur_fund_ids) - c.proper_fund_num]:
                    to_cancel.add(fund_id)
                    print(f'to_cancel (fund_id){fund_id} (rank){_rank}')

            for fund_id in c.cur_fund_ids:
                _rank = c.fund_ranks.get(fund_id)
                if _rank is None:
                    print(f'got a None rank when trying to trade, (index_id){c.index_id} (fund_id){fund_id} (dt){dt} (fund_scores){c.fund_scores} (fund_ranks){c.fund_ranks}')
                    to_cancel.add(fund_id)
                _cur_wgt = c.fund_cur_wgts.get(fund_id, 0)
                _tar_wgt = 0
                if fund_id in to_cancel:
                    _tar_wgt = 0
                elif _rank <= c.proper_fund_num:
                    # 当前持仓仍排名在前 proper_num 的基金，我们统一安排 1 / proper_num 的比例
                    _tar_wgt = proper_fund_wgt
                    if _cur_wgt * c.index_cur_wgt > _tar_wgt * c.index_tar_wgt:
                        good_fund_to_sell.append((fund_id, _rank, _cur_wgt * c.index_cur_wgt - _tar_wgt * c.index_tar_wgt))
                elif _rank <= c.proper_fund_num * 2:
                    # 如果排名在 proper_num + 1 ~ 2 * proper_num 的基金，
                    # 如果不足 1 / proper_num 的比例，则不变，如果多于 1 / proper_num，就减至 1 / proper_num。（只减少，不增加）
                    _tar_wgt = min(proper_fund_wgt, _cur_wgt * c.index_cur_wgt / c.index_tar_wgt)
                else:
                    _tar_wgt = 0

                fund_selected_wgt += _tar_wgt
                fund_selected_num += 1 if _tar_wgt > 0 else 0

                _wgt = FundWeightItem(fund_id=fund_id, index_id=c.index_id, asset_wgt=c.index_tar_wgt, fund_wgt_in_asset=_tar_wgt)
                funds[fund_id] = _wgt

            assert fund_selected_wgt <= 1, f'[fund trade] tot fund wgt <= 1: {fund_selected_wgt}'

            # 如果当前比例还有剩余（步骤2中产生的）
            if fund_selected_wgt < 1 - 1e-5 and c.index_tar_wgt > 0:
                tot_wgt_left = (1 - fund_selected_wgt) * c.index_tar_wgt
                # 如果步骤1中前 proper_num 有哪个减少过，就按照排名高低revert这个减少
                if len(good_fund_to_sell) > 0:
                    good_to_sell_list = sorted(good_fund_to_sell, key=lambda x: x[1])

                    for fund_id, _rank, _sell_wgt in good_to_sell_list:
                        print(f'one in good_to_sell_list, (index_id){c.index_id} (fund_id){fund_id} (rank){_rank} (sell_wgt){_sell_wgt}')
                        revert_wgt = min(tot_wgt_left, _sell_wgt)
                        funds[fund_id].add_tot_wgt(revert_wgt)
                        tot_wgt_left -= revert_wgt
                        if tot_wgt_left < 1e-5:
                            break

                # 如果将减少的都处理完还有剩余，那么进入到特殊状态（SPECIAL）
                if tot_wgt_left > 1e-5:
                    print(f"happy in SPECIAL CASE (proper num){c.proper_fund_num}")
                    assert fund_selected_num <= c.proper_fund_num, f'fund selected num {fund_selected_num} should le than proper fund num {c.proper_fund_num}!!'
                    # 对于SPECIAL，我们先看fund_selected_num是否已达到proper_num
                    if fund_selected_num < c.proper_fund_num:
                        # 如果否，从fund_score中按评分从高到低将不在当前持仓的基金每个都增加进去，最多不超过 1 / proper_num，直到fund_selected_num等于proper_num或剩余的量消耗光
                        for fund_id, _score in c.fund_scores:
                            if fund_id not in funds:
                                _tar_wgt = min(proper_fund_wgt, tot_wgt_left / c.index_tar_wgt)
                                _wgt = FundWeightItem(fund_id=fund_id, index_id=c.index_id, asset_wgt=c.index_tar_wgt, fund_wgt_in_asset=_tar_wgt)
                                print(f'[SPECIAL] (index_id){c.index_id} (fund){fund_id} (score){_score} (rank){c.fund_ranks[fund_id]} (tar_wgt){_tar_wgt}')
                                funds[fund_id] = _wgt
                                fund_selected_num += 1
                                tot_wgt_left -= _tar_wgt * c.index_tar_wgt
                                if tot_wgt_left < 1e-5 or fund_selected_num == c.proper_fund_num:
                                    break

                    # 如果是，或fund_selected_num等于proper_num时剩余的量仍未消耗光
                    # 我们将剩余的比例平均分配给当前持仓的基金，以此来将剩余的量消耗光（分配后有可能超过 1 / proper_num，但实际在上边第4条中也有可能导致超过 1 / proper_num）
                    if tot_wgt_left > 1e-5:
                        funds_without_no_wgt = {fund_id: fund for fund_id, fund in funds.items() if fund.fund_wgt_in_asset > 0}
                        avg_wgt_left = tot_wgt_left / len(funds_without_no_wgt)
                        for fund_id, _wgt in funds_without_no_wgt.items():
                            print(f'[SPECIAL] (index_id){c.index_id} (fund){fund_id} (rank){c.fund_ranks[fund_id] if fund_id in c.fund_ranks else None} (avg_wgt_left){avg_wgt_left}')
                            _wgt.add_tot_wgt(avg_wgt_left)
                            tot_wgt_left -= avg_wgt_left
                            assert tot_wgt_left >= 0 or math.isclose(tot_wgt_left, 0, abs_tol=1e-05), f'tot_wgt_left less than zero!! {tot_wgt_left} (index_id){c.index_id}'

                assert tot_wgt_left < 1e-5, f'special even failed: {c.index_id}'

            for fund_id, _wgt in funds.items():
                target_fund_amt = _wgt.fund_wgt * fund_tot_mv
                cur_fund_amt = c.fund_cur_wgts.get(fund_id, 0) * c.index_cur_wgt * fund_tot_mv
                if abs(target_fund_amt - cur_fund_amt) > fund_tot_mv * self.fund_param.MinCountAmtDiff or (target_fund_amt == 0 and cur_fund_amt > 0):
                    p = cur_fund_nav[fund_id]
                    _trade = FundTrade(
                        fund_id=fund_id,
                        index_id=c.index_id,
                        mark_price=p,
                        amount=abs(target_fund_amt - cur_fund_amt),
                        is_buy=target_fund_amt > cur_fund_amt,
                        submit_date=dt,
                        is_to_cleanup=(target_fund_amt == 0 and cur_fund_amt > 0)
                    )
                    fund_trades.append(_trade)
                    print(f'(fund){fund_id} (index){c.index_id} (p){p} (rank){c.fund_ranks[fund_id] if fund_id in c.fund_ranks else 0} (cur_amt){cur_fund_amt} (tar_amt){target_fund_amt} (idx){c.index_id} (amt){target_fund_amt - cur_fund_amt}')

        return fund_trades


import pandas as pd
from ....manager.data_tables import StockInfoDataTables


def unify_dates(sdt: StockInfoDataTables):
    finance = sdt.finance
    finance = pd.concat([finance, sdt.real_dates], axis=1)
    finance = finance.drop_duplicates(subset=['stmt_act_date'], keep='last')
    finance = finance.set_index('stmt_act_date')
    finance = finance.dropna(axis=0, how='all')

    equity = sdt.equity.reindex(sdt.prices.index)
    df = pd.concat([finance, sdt.prices, equity, sdt.dividend], axis=1)
    df['日涨跌'] = df['收盘价'] / df['收盘价'].shift(1) - 1
    df = df.sort_index()
    df = df.fillna(method='pad')

    return df


def compute_value_ratios(sdt: StockInfoDataTables):
    df = unify_dates(sdt)

    df['总市值'] = df['收盘价'] * df['总股本']
    df['年度派息总额'] = df['每股股息'] * df['总股本']
    df['TTM.市盈率'] = df['总市值'] / df['TTM.归属于母公司股东的净利润']
    df['市净率'] = df['总市值'] / df['净资产']
    df['TTM.市销率'] = df['总市值'] / df['TTM.营业总收入']
    df['TTM.市现率'] = df['总市值'] / df['TTM.经营活动产生的现金流量净额']
    df['市息率'] = df['总市值'] / df['年度派息总额']
    df['股息率'] = df['年度派息总额'] / df['总市值']
    df['TTM.派息率'] = df['年度派息总额'] / df['TTM.归属于母公司股东的净利润']

    df['每股净资产'] = df['净资产'] / df['总股本']

    df['PEG'] = df['TTM.市盈率'] / df['近1年净利润增速']
    return df[[
        'TTM.市盈率',
        '市净率',
        'TTM.市销率',
        'TTM.市现率',
        '市息率',
        '股息率',
        'TTM.派息率',
        '每股净资产',
        'PEG',
        '收盘价',
        '日涨跌',
        '总市值',
        '年度派息总额',
        '总股本',
        'TTM.归属于母公司股东的净利润',
        '净资产',
        'TTM.营业总收入',
        'TTM.经营活动产生的现金流量净额',
    ]]





import json
import os
import traceback
from typing import Dict, List

import pandas as pd
import numpy as np

from .....constant import SourceType
from .....fund.engine.trader import BasicFundTrader
from .....fund.engine.backtest import BasicFundBacktestEngine
from ....fund.basic.basic_data_helper import BasicDataHelper
from ....manager.manager_fund import FundDataManager
from ....view.derived_models import ThirdPartyPortfolioInfo, ThirdPartyPortfolioTrade, ThirdPartyPortfolioIndicator, ThirdPartyPortfolioPositionLatest
from ..derived_data_helper import DerivedDataHelper, DateJSONEncoder
from .common import ThirdPortfolioBase


class GFSecPortfolioProcessor(ThirdPortfolioBase):

    _BLOCK_LIST = [
    ]

    _FILE_SUFFIX = 'json'
    _INFO_COLUMNS_MAP = {
        'name': 'po_name',
        'setupDate': 'start_date',
    }

    def __init__(self, data_helper: DerivedDataHelper):
        super().__init__()

        self._data_helper = DerivedDataHelper()
        self._params_json: Dict[str, str] = {}
        self._dir_path = f'{os.path.dirname(os.path.abspath(__file__))}/data/gf_sec'
        self._trades_df_list: List[pd.DataFrame] = []
        self._nav: Dict[str, pd.DataFrame] = {}
        self._last_position: List[pd.Series] = []

    def init(self):
        # self._fdm = FundDataManager(s3_retriever_activation=True)
        # self._fdm.init(score_pre_calc=True, print_time=True)

        from .....resource.data_store import DataStore
        self._fdm: FundDataManager = DataStore.load_dm()
        self._bdh = BasicDataHelper()
        self._trading_day: pd.DataFrame = FundDataManager.basic_data('get_trading_day_list').drop(columns='_update_time')

        self._list_all = pd.read_csv(f'{self._dir_path}/list_all.txt', header=None, names=['file_name'])

    def _calc_basic_info(self, x: pd.Series) -> str:
        with open(f'{self._dir_path}/portfolios/{x.file_name}', 'r') as f:
            portfolio_base_info = json.load(f)

        base_info = portfolio_base_info.get('data')
        if base_info is None:
            print(f'[WARNING] Can not find info of {x.file_name}')
            return

        po_id = x.file_name.split('.')[0]
        assert po_id == base_info.get('advId'), 'portfolio id does not match'

        result = {}
        result['po_id'] = po_id
        for one in ('name', 'setupDate'):
            result[one] = base_info.get(one)
        introduce_info = base_info.get('investAdvNodeForPlanShow')
        if introduce_info is not None:
            result['po_desc'] = introduce_info.get('INTRODUCE')
            result['rich_desc'] = introduce_info.get('STRATEGY')
            result['risk_level'] = introduce_info.get('INCOMESTYLE')
        return pd.Series(result)

    def _calc_trades(self, x: pd.Series):
        po_id = x.file_name.split('.')[0]

        with open(f'{self._dir_path}/portfolios/{x.file_name}', 'r') as f:
            portfolio_base_info = json.load(f)
        base_info = portfolio_base_info.get('data')
        if base_info is None:
            print(f'[WARNING] Can not find base info of {po_id} (po_base_info){portfolio_base_info}')
            return
        plan_date_list = base_info.get('planDateList')
        if plan_date_list is None:
            print(f'[WARNING] Can not find plan date list of {po_id} (base_info){base_info}')
            return
        plan_date_list = set(plan_date_list)

        can_not_calc = False
        fund_weight_list = []
        result_params = {}
        result_params['fund_weight_list'] = []
        with open(f'{self._dir_path}/portfolios/{po_id}_adj.{self._FILE_SUFFIX}', 'r') as f:
            adj_info = json.load(f)
        for date, value in adj_info.items():
            assert date in plan_date_list, f'{date} not in the plan date list (po_id){po_id}'

            trans_data = value.get('data')
            if trans_data is None:
                print(f'[WARNING] Invalid trans data {value} (po_id){po_id}')
                continue
            assert trans_data.get('advId') == po_id, 'portfolio id does not match'
            assert trans_data.get('updateDate') == date, 'portfolio trans date does not match'
            comment = trans_data.get('transReason')
            type_list = trans_data.get('typeList')
            if type_list is None:
                print(f'[WARNING] Invalid type list {trans_data} (po_id){po_id}')
                continue
            trade_date = pd.to_datetime(date, infer_datetime_format=True).date()
            fp_sum = 0
            tp_sum = 0
            result_details = []
            for one_item in type_list:
                fund_list = one_item.get('fundList')
                if fund_list is None:
                    print(f'[WARNING] Invalid fund list {one_item} (po_id){po_id}')
                    continue
                if not fund_list:
                    print(f'[INFO] Empty fund list, do next one')
                    continue
                for one in fund_list:
                    raw_fund_id = one.get('fundCode')
                    fund_id = self._bdh._get_fund_id_from_order_book_id(raw_fund_id, trade_date)
                    if fund_id is None:
                        print(f'[WARNING] Can not found fund {raw_fund_id} (po_id){po_id}')
                        fund_id = raw_fund_id + '!0'
                    from_percent = one.get('prePercentStr')
                    to_percent = one.get('percentStr')
                    if fund_id in self._BLOCK_LIST:
                        can_not_calc = True
                    result_details.append({'fund_id': fund_id, 'from_percent': float(from_percent) if from_percent else 0, 'to_percent': float(to_percent) if to_percent else 0})
                    fp_sum += float(from_percent) if from_percent else 0
                    tp_sum += float(to_percent) if to_percent else 0
            fp_sum = round(fp_sum, 3)
            tp_sum = round(tp_sum, 3)
            if (fp_sum != 0 and fp_sum != 1) or tp_sum != 1:
                print(f'sum of percent not equals 1 (po_id){po_id} (trade_date){trade_date} (fp_sum){fp_sum} (tp_sum){tp_sum}')
            adjusted_date = trade_date
            # 我们是第二天调仓，因此这里调整为之前的第一个交易日
            adjusted_date = self._trading_day[self._trading_day.datetime < adjusted_date].datetime.array[-1]
            fund_weight_list.append({'po_id': po_id, 'datetime': trade_date, 'details': json.dumps(result_details), 'comment': comment})
            result_params['fund_weight_list'].append([adjusted_date, [(one['fund_id'], one['to_percent']) for one in result_details]])
        self._trades_df_list.append(pd.DataFrame(fund_weight_list))
        if can_not_calc:
            return

        result_params['start_date'] = base_info.get('setupDate')
        result_params['end_date'] = base_info.get('netvalue_date')
        result_params['fund_weight_list'].sort(key=lambda x: x[0])
        self._params_json[po_id] = json.dumps(result_params, indent=4, cls=DateJSONEncoder)

    def _calc_indicators(self, x: pd.Series):
        po_id = x.file_name.split('.')[0]
        params_json = self._params_json.get(po_id)
        if params_json is None:
            return
        params = json.loads(params_json)

        start_date = pd.to_datetime(params.get('start_date'), infer_datetime_format=True).date()
        # 我们是第二天调仓，因此这里调整为之前的第一个交易日
        start_date = self._trading_day[self._trading_day.datetime < start_date].datetime.array[-1]
        # end_date = pd.to_datetime(params.get('end_date'), infer_datetime_format=True).date()

        t = BasicFundTrader()
        b = BasicFundBacktestEngine(data_manager=self._fdm, trader=t, fund_weight_list=params['fund_weight_list'])
        b.init()
        b.run(start_date=start_date, end_date=None, print_time=False)

        last_position: pd.DataFrame = b.get_last_position()
        self._last_position.append(pd.Series({'po_id': po_id, 'datetime': last_position.index.name, 'po_src': SourceType.gf_sec, 'position': last_position.loc[:, 'weight'].to_json()}))

        result = b.get_fund_result()
        # 只保留我们需要的回测结果
        result = {k: v for k, v in result.items() if k in ('market_value', 'last_unit_nav', 'mdd', 'annual_ret', 'mdd_d1', 'mdd_d2', 'total_commission', 'rebalance_times')}
        result['datetime'] = result['market_value'].index.array[-1]
        result['from_setup_ret'] = result['last_unit_nav'] - 1
        resampled = result['market_value'].copy()
        resampled.index = pd.to_datetime(resampled.index)
        resampled_w = resampled.mv.resample('1W').apply(lambda x: x[-1] if not x.empty else np.nan)
        result['vol_by_week'] = resampled_w.dropna().pct_change().std(ddof=1) * np.sqrt(52)

        resampled_m = resampled.mv.resample('1M').apply(lambda x: x[-1] if not x.empty else np.nan)
        result['vol_by_month'] = resampled_m.dropna().pct_change().std(ddof=1) * np.sqrt(12)
        result['vol_by_day'] = resampled.mv.dropna().pct_change().std(ddof=1) * np.sqrt(244)
        result['annual_compounded_ret'] = np.power(result['last_unit_nav'], 1 / (result['market_value'].shape[0] / 244)) - 1
        if result['vol_by_week'] != 0:
            result['sharpe_ratio'] = (result['annual_compounded_ret'] - 0.02) / result['vol_by_week']
        else:
            result['sharpe_ratio'] = np.nan

        # 遍历每个需要计算的时间区间
        for name, value in self._DIM_TYPE.items():
            bd = (result['datetime'] - value).date()
            # 试图寻找bd及其之前的最近一天
            filtered_mv = result['market_value'].loc[result['market_value'].index <= bd, :]
            if filtered_mv.empty:
                continue
            # 计算区间内的return
            result[name] = result['market_value'].mv.array[-1] / filtered_mv.mv.array[-1] - 1

        result['nav'] = result['last_unit_nav']
        result['po_id'] = po_id
        self._nav[po_id] = result['market_value'].mv / result['market_value'].mv.array[0]
        self._nav[po_id].index = self._nav[po_id].index.map(lambda x: x.isoformat())
        # 删掉我们不需要的回测结果
        for one in ('market_value', 'last_unit_nav'):
            del result[one]
        print(f'{po_id} done')
        return pd.Series(result)

    def calc(self):
        basic_info: pd.DataFrame = self._list_all.apply(self._calc_basic_info, axis=1)
        basic_info['po_src'] = SourceType.gf_sec
        basic_info['manager_name'] = '广发证券'
        self._data_helper._upload_derived(basic_info.rename(columns=self._INFO_COLUMNS_MAP), ThirdPartyPortfolioInfo.__table__.name)

        self._list_all.apply(self._calc_trades, axis=1)
        trades = pd.concat(self._trades_df_list)
        trades['po_src'] = SourceType.gf_sec
        self._data_helper._upload_derived(trades, ThirdPartyPortfolioTrade.__table__.name)

        indicators: pd.DataFrame = self._list_all.apply(self._calc_indicators, axis=1)
        indicators = indicators[indicators.notna().any(axis=1)]
        indicators['po_src'] = SourceType.gf_sec
        self._data_helper._upload_derived(indicators, ThirdPartyPortfolioIndicator.__table__.name)

        last_position = pd.DataFrame(self._last_position)
        self._data_helper._upload_derived(last_position, ThirdPartyPortfolioPositionLatest.__table__.name)

        # nav存cas
        from ....view.cas.third_party_portfolio_nav import ThirdPartyPortfolioNav
        # from cassandra.cqlengine.query import BatchQuery

        # with BatchQuery() as batch:
        #     for po_id, mv in self._nav.items():
        #         ThirdPartyPortfolioNav.objects(po_id=po_id).batch(batch).update(
        #             date__append=mv.index.to_list(),
        #             nav__append=mv.to_list(),
        #         )
        for po_id, mv in self._nav.items():
            ThirdPartyPortfolioNav.create(
                po_id=po_id,
                date=mv.index.to_list(),
                nav=mv.to_list(),
            )

    def process(self, end_date: str) -> List[str]:
        failed_tasks: List[str] = []
        try:
            self.calc_everyday(end_date, [SourceType.gf_sec.value])
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('third_party_portfolio_gf_sec')
        return failed_tasks


if __name__ == '__main__':
    gf_sec_pp = GFSecPortfolioProcessor(DerivedDataHelper())
    # gf_sec_pp.init()
    # gf_sec_pp.calc()
    # gf_sec_pp.process('20201012')



import pandas as pd
import numpy as np
import os
import datetime
import traceback
import requests
import urllib3
from ...wrapper.mysql import RawDatabaseConnector
from .raw_data_helper import RawDataHelper
from ...view.raw_models import SinaStockMinute


class SinaStockDownloader(object):

    def __init__(self, data_helper, em_stock_list):
        req_batch_count = 5
        self.url_prefix = 'http://hq.sinajs.cn/list='
        self.headers = {'User-Agent': ('Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/605.1.15 '
                                      '(KHTML, like Gecko) Version/13.0.5 Safari/605.1.15')}

        self._data_helper = data_helper
        self.sina_stock_list = self.parse_em_stock_list(em_stock_list)
        self.sina_stock_req_batch = self.parse_sina_stock_req_batch(
            self.sina_stock_list, req_batch_count)

    def parse_em_stock_list(self, em_stock_list):
        sina_stock_list = []
        for em_stock_id in em_stock_list:
            sina_stock_id = self.em_stock_id_to_sina(em_stock_id)
            if sina_stock_id:
                sina_stock_list.append(sina_stock_id)

        return sina_stock_list

    def parse_sina_stock_req_batch(self, sina_stock_list, req_batch_count):
        sina_stock_req_content_list = []
        total_count = len(sina_stock_list)
        if req_batch_count < 1 or total_count < req_batch_count:
            print(f'Invalid stock count({total_count}) or batch count({req_batch_count})')
            return None

        req_batch_size = int(total_count / req_batch_count)
        start = 0
        for batch_index in range(req_batch_count):
            if batch_index == req_batch_count - 1:
                end = total_count
            else:
                end = start + req_batch_size
            sina_stock_req_batch_content = ','.join(
                sina_stock_list[start: end])
            start += req_batch_size
            sina_stock_req_content_list.append(sina_stock_req_batch_content)

        return sina_stock_req_content_list

    def sina_stock_id_to_em(self, sina_stock_id):
        if sina_stock_id.startswith('sh'):
            return f'{sina_stock_id[2:]}.SH'
        elif sina_stock_id.startswith('sz'):
            return f'{sina_stock_id[2:]}.SZ'
        else:
            print(f'Unrecognized sina stock code {sina_stock_id}')
            return None

    def em_stock_id_to_sina(self, em_stock_id):
        parts = em_stock_id.split('.')
        if len(parts) != 2:
            print(f'Unrecognized stock code {em_stock_id}')
            return None

        if parts[1] == 'SH':
            return f'sh{parts[0]}'
        elif parts[1] == 'SZ':
            return f'sz{parts[0]}'
        else:
            print(f'Unrecognized stock code {em_stock_id}')
            return None

    def parse_rsp_line(self, line):
        '''
        Below line is a sample:
        var hq_str_sh600687="*ST刚泰,1.360,1.340,1.410,1.410,1.320,1.410,0.000,25705789,35407399.000,86885,1.410,537100,1.400,316000,1.390,271900,1.380,187200,1.370,0,0.000,0,0.000,0,0.000,0,0.000,0,0.000,2020-08-06,15:00:01,00,";
        '''
        if not line:
            return None

        parts = line.split('=')
        if len(parts) != 2:
            print(f'Data error: {line}')
            return None
        sina_stock_id = parts[0].split('_')[-1]
        em_stock_id = self.sina_stock_id_to_em(sina_stock_id)

        right_parts = parts[1].split(',')
        if len(right_parts) < 33:
            print(f'Data error: {line}')
            return None

        if right_parts[32].endswith('";'):
            right_parts[32] = right_parts[32][:-2]

        if right_parts[32] != '00' and right_parts[32] != '00':
            print(f'Status of stock {em_stock_id} is not valid({right_parts[32]})')
            return None

        price = float(right_parts[3])
        dt = datetime.datetime.strptime(f'{right_parts[30]} {right_parts[31]}', '%Y-%m-%d %H:%M:%S')
        
        return [em_stock_id, price, dt]

    def sina_stock_minute(self, to_save):
        try:
            stock_prices = []
            for req_content in self.sina_stock_req_batch:
                url = f'{self.url_prefix}{req_content}'

                r = requests.get(url, headers=self.headers)
                lines = r.text.split('\n')
                for line in lines:
                    result = self.parse_rsp_line(line)
                    if result:
                        stock_prices.append(result)

            df = pd.DataFrame(stock_prices, columns=['stock_id', 'close', 'datetime'])
            if to_save:
                try:
                    self._data_helper._upload_raw(df, SinaStockMinute.__tablename__)
                except Exception as e:
                    print(e)
                    traceback.print_exc()
            return df
        except Exception as e:
            print(e)
            traceback.print_exc()
            return None




from typing import Optional

import pandas as pd

from .derived import DerivedFactor
from ...constant import StockFactorType


class MixFactorBase(DerivedFactor):

    def get(self, universe: str = 'default') -> Optional[pd.DataFrame]:
        if universe not in self._factor:
            try:
                self._factor[universe]: pd.DataFrame = pd.read_parquet(self._get_s3_factor_uri_with_universe(universe))
            except Exception as e:
                print(f'retrieve data from s3 failed, (err_msg){e}; try to re-calc')
                self.calc(universe)
        return self._factor[universe]

    def save(self, universe: str = 'default') -> bool:
        if self._factor is None or universe not in self._factor:
            return False
        self._factor[universe].to_parquet(self._get_s3_factor_uri_with_universe(universe), compression='gzip')
        return True


class ValueFactor(DerivedFactor):
    # 所有价值因子的基类
    def __init__(self, name):
        super().__init__(f'val/{name}', StockFactorType.VALUE)


class GrowthFactor(DerivedFactor):
    # 所有成长因子的基类
    def __init__(self, name):
        super().__init__(f'gro/{name}', StockFactorType.GROWTH)


class FinQualityFactor(DerivedFactor):
    # 所有财务质量因子的基类
    def __init__(self, name):
        super().__init__(f'fin/{name}', StockFactorType.FIN_QUALITY)


class VolatilityFactor(DerivedFactor):
    # 所有波动率因子的基类
    def __init__(self, name):
        super().__init__(f'vol/{name}', StockFactorType.VOLATILITY)


class ScaleFactor(DerivedFactor):
    # 所有规模因子的基类
    def __init__(self, name):
        super().__init__(f'sca/{name}', StockFactorType.SCALE)


class LeverageFactor(DerivedFactor):
    # 所有杠杆因子的基类
    def __init__(self, name):
        super().__init__(f'lev/{name}', StockFactorType.LEVERAGE)


class MomentumFactor(DerivedFactor):
    # 所有动量因子的基类
    def __init__(self, name):
        super().__init__(f'mom/{name}', StockFactorType.MOMENTUM)


class LiquidityFactor(DerivedFactor):
    # 所有流动性因子的基类
    def __init__(self, name):
        super().__init__(f'liq/{name}', StockFactorType.LIQUIDITY)


class TechFactor(DerivedFactor):
    # 所有技术因子的基类
    def __init__(self, name):
        super().__init__(f'tech/{name}', StockFactorType.TECH)


class Alpha101Factor(DerivedFactor):
    # 所有alpha101因子的基类
    def __init__(self, name):
        super().__init__(f'alpha101/{name}', StockFactorType.ALPHA101)



from typing import Tuple, List

import pandas as pd
import datetime
from sqlalchemy import func
import numpy as np
from .raw import RawDataApi
from .basic import BasicDataApi
from .derived import DerivedDataApi
from ...util.singleton import Singleton
from ...util.calculator_item import CalculatorBase
from ...util.calculator import Calculator

class PortfolioDataApi(metaclass=Singleton):

    def get_portfolio_nav(self, order_list, fund_type):
        try:
            share_list = {}
            port_amount = 1
            result = []
            group_result = []
            for i in range(len(order_list)):
                trade_date = list(order_list[i].keys())[0]
                if i + 1 < len(order_list):
                    change_date = list(order_list[i + 1].keys())[0]
                else :
                    change_date = datetime.date.today()
                fund_list = list(list(order_list[i].values())[0].keys())
                fund_nav = BasicDataApi().get_fund_nav_with_date(start_date=trade_date,end_date=change_date,fund_list=fund_list)
                fund_nav = fund_nav.pivot_table(index="datetime" , values="adjusted_net_value" , columns="fund_id").dropna()
                dict_share = {}
                for j in range(len(fund_list)):
                    fund_share = port_amount * order_list[i][trade_date][fund_list[j]] / fund_nav.iloc[0][fund_list[j]]
                    dict_share[fund_list[j]] = fund_share
                    fund_nav[fund_list[j]] = fund_nav[fund_list[j]] * fund_share
                
                _result = []
                for index_group, fund_list in fund_type.items():
                    ws = sum([order_list[i][trade_date].get(fund_id,0) for fund_id in fund_list])
                    nav_mul = 1/ws
                    _df = (fund_nav[fund_nav.columns.intersection(fund_list)]*nav_mul).sum(axis=1).to_frame()
                    _df.columns=[index_group]
                    _result.append(_df)
                group_result.append(pd.concat(_result,axis=1))
                share_list[trade_date] =  dict_share 
                port_amount = sum(fund_nav.iloc[-1])
                if i == 0 : 
                    result.append(fund_nav)
                    fund_nav = fund_nav.sum(axis = 1).to_frame()
                    fund_nav.columns = ["组合"]
                    port_mv = fund_nav
                else :
                    result.append(fund_nav)
                    fund_nav = fund_nav.sum(axis = 1).to_frame()
                    fund_nav.columns = ["组合"]
                    port_mv = port_mv.append(fund_nav.iloc[1:])
            port_fund_nav = pd.concat(result)
            port_fund_weight = port_fund_nav.div(port_fund_nav.sum(axis=1), axis=0)
            port_fund_group = pd.concat(group_result)
            data = {
                '净值':port_mv,
                '各基净值':port_fund_nav,
                '各基权重':port_fund_weight,
                '大类净值':port_fund_group,
                '持有信息':share_list,
            }
            return data
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_portfolio_mv')


    def get_portfolio_benchmark_info(self):
        try:
            index_info = BasicDataApi().get_index_info(index_list=['hs300','national_debt','csi500','gem','sp500rmb','mmf','hsi'])
            res = []
            for r in index_info.itertuples():
                res.append([r.index_id, r.desc_name])
            return res

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_portfolio_benchmark_info')

    def get_portfolio_mv(self, fund_nav, index_id, begin_date, end_date, time_para):
        try:
            fund_nav.columns=['组合']
            basic = BasicDataApi()
            index_price = basic.get_index_price_dt(index_list=[index_id],start_date=fund_nav.index[0] - datetime.timedelta(days=10),end_date=fund_nav.index[-1]).pivot_table(index='datetime',columns='index_id',values='close')
            index_info = basic.get_index_info([index_id])
            name_dic = index_info.set_index('index_id').to_dict()['desc_name']
            begin_date, end_date = RawDataApi().get_date_range(time_para, begin_date, end_date)
            df = pd.merge(fund_nav.reset_index(), index_price.reset_index(),how='outer').set_index('datetime').sort_index().ffill().dropna()
            df = df.loc[begin_date:end_date]
            df = df / df.iloc[0]
            df.loc[:,'价格比'] = df.组合 / df[index_id]
            df = df.rename(columns=name_dic)
            return df
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_portfolio_mv')

    def get_portfolio_cur_mdd(self, fund_nav, begin_date, end_date, time_para):
        try:            
            begin_date, end_date = RawDataApi().get_date_range(time_para, begin_date, end_date)
            fund_nav = fund_nav.loc[begin_date:end_date]
            df = fund_nav / fund_nav.cummax() - 1
            df.columns=['历史回撤']
            df_min = df['历史回撤'].min()
            data = {
                '历史回撤': df,
                '最低值': df_min,
            }
            return data
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_portfolio_cur_mdd')

    def get_portfolio_ret_distribution(self, fund_nav, period, begin_date, end_date, time_para):
        try:
            begin_date, end_date = RawDataApi().get_date_range(time_para, begin_date, end_date)
            fund_nav = fund_nav.loc[begin_date:end_date]    
            fund_nav = CalculatorBase.data_resample_weekly_nav(df=fund_nav, rule=period)
            ret = fund_nav.pct_change(1).iloc[1:]
            ret.columns=['收益']
            result = np.histogram(a=ret['收益'],bins=8)
            data = {'分布_频度':result[0].tolist(),
                '分布_收益':result[1].round(3).tolist(),
                '收益':ret
            }
            return data
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_portfolio_cur_mdd')

    def get_portfolio_stats(self, fund_nav, index_id):
        try:
            fund_nav.columns=['组合']
            basic = BasicDataApi()
            index_price = basic.get_index_price_dt(index_list=[index_id],start_date=fund_nav.index[0] - datetime.timedelta(days=10),end_date=fund_nav.index[-1]).pivot_table(index='datetime',columns='index_id',values='close')
            index_info = basic.get_index_info([index_id])
            name_dic = index_info.set_index('index_id').to_dict()['desc_name']
            df = pd.merge(fund_nav.reset_index(), index_price.reset_index(),how='outer').set_index('datetime').sort_index().ffill().dropna()
            df = df / df.iloc[0]
            df = df.rename(columns=name_dic)
            result = []
            for col in df:
                res_status = Calculator.get_benchmark_stat_result(dates=df.index,
                                                                values=df[col],
                                                                benchmark_values=df[name_dic[index_id]],
                                                                frequency = '1W',
                                                                risk_free_rate=0.025, )
                
                indicators_tem={
                    '开始日期': res_status.start_date,
                    '截止日期': res_status.end_date,
                    '累计收益': str(round((res_status.last_unit_nav-1)*100,2)) + "%",
                    '最近一月收益':str(round(res_status.recent_1m_ret*100 ,2)) + "%",
                    '年化收益率': str(round(res_status.annualized_ret*100 ,2)) + "%",
                    '年化波动率': str(round(res_status.annualized_vol*100 ,2)) + "%",
                    '夏普比率': round(res_status.sharpe,2),
                    '最大回撤': str(round(res_status.mdd*100 ,2)) + "%",
                    '最大回撤开始日期': res_status.mdd_date1,
                    '最大回撤结束日期': res_status.mdd_date2,
                    '最大回撤持续时长': res_status.mdd_lens,
                    '卡玛比率': round(res_status.ret_over_mdd ,2),
                    '下行标准差': str(round(res_status.downside_std*100 ,2)) + "%",
                    'alpha': str(round(res_status.alpha*100 ,2)) + "%",
                    'beta': round(res_status.beta ,2),
                    '信息比率': round(res_status.ir ,2),
                    'CL模型_alpha': str(round(res_status.alpha_cl*100 ,2)) + "%",
                    'CL模型_beta': str(round(res_status.beta_cl*100 ,2)) + "%",
                    '相对胜率': str(round(res_status.win_rate*100 ,2)) + "%",
                    '绝对胜率': str(round(res_status.win_rate_0*100 ,2)) + "%",
                    'name':col
                }
                if col != '组合':
                    indicators_tem['alpha'] = '-'
                    indicators_tem['beta'] = '-'
                    indicators_tem['信息比率'] = '-'
                    indicators_tem['CL模型_alpha'] = '-'
                    indicators_tem['CL模型_beta'] = '-'
                    indicators_tem['相对胜率'] = '-'
                result.append(indicators_tem)
            df = pd.DataFrame(result).set_index('name').T
            return df
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_portfolio_stats')

    def get_rolling_corr(self, fund_navs, frequency, windows, step):
        try:
            fund_navs_period = CalculatorBase.data_resample_weekly_nav(df=fund_navs, rule=frequency)
            fund_ret = fund_navs_period.pct_change(1).iloc[1:]  
            # 弱鸡排列
            col_pair = []
            cols = fund_ret.columns.tolist()
            for idx_i, col_i in enumerate(cols):
                cols_list = cols[idx_i+1:]
                for idx_j, col_j in enumerate(cols_list):
                    col_pair.append([col_i,col_j])

            corr_result_df = []
            for col_pair_i in col_pair:
                df = fund_ret[col_pair_i]
                idx_list = df.reset_index().index.tolist()
                corr_result = []
                b = idx_list[0]
                _result = []
                
                # 计算窗口划分 不包含min_windows概念
                while(1):
                    e = b + windows
                    if e > idx_list[-1]:
                        break
                    else:
                        _result.append([b,e])
                        b += step

                # 计算部分
                for idxs in _result:
                    _df = df.iloc[idxs[0]:idxs[1]]
                    corr_result.append({'datetime':_df.index[-1], f'{col_pair_i[0]}_{col_pair_i[1]}':_df.corr().iloc[1,0]})
                df = pd.DataFrame(corr_result).set_index('datetime')
                corr_result_df.append(df)
            df = pd.concat(corr_result_df,axis=1)
            return df
        
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_rolling_corr')
    
    def get_port_fund_list(self, order_list):
        try:
            fund_list = []
            for order in order_list:
                for d, fund_wgt in order.items():
                    fund_list.extend(list(fund_wgt))
            return list(set(fund_list))
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_port_fund_list')

    def get_port_fund_pos_date(self,fund_id, order_list,index_id): 
        try:
            basic = BasicDataApi()
            begin_date, end_date = RawDataApi().get_date_range(time_para='ALL', begin_date=None, end_date=None)
            fund_nav = basic.get_fund_nav(fund_list=[fund_id])
            fund_nav = fund_nav.pivot_table(columns='fund_id',index='datetime',values='adjusted_net_value')
            index_price = basic.get_index_price(index_list=[index_id])
            index_price = index_price.pivot_table(columns='index_id',index='datetime',values='close')
            nav_df = fund_nav.join(index_price).ffill().dropna()
            
            result = []
            for order in order_list:
                for d, fund_wgt in order.items():
                    fund_wgt['datetime'] = d
                    result.append(fund_wgt)

            weight_df = pd.DataFrame(result).set_index('datetime')
            pos_con = 1 - weight_df[fund_id].isnull()
            result = []
            is_empty = True
            for r in pos_con.iteritems():
                if is_empty and r[1] == 1:
                    dic = {'begin_date':r[0]}
                    is_empty = False
                elif not is_empty and r[1] == 0:
                    dic['end_date'] = r[0]
                    result.append(dic)
                    dic = {}
                    is_empty = True
            if not is_empty:
                dic['end_date'] = end_date
                result.append(dic)
            data = {
                '净值':nav_df,
                '持有期':result}
            return data
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_port_fund_pos_date')

    def get_port_ret_resolve_date_range(self, order_list):
        try:
            
            begin_date, end_date = RawDataApi().get_date_range(time_para='ALL', begin_date=None, end_date=None)
            begin_date = list(order_list[0].keys())[0]
            data = {'begin_date':begin_date,'end_date':end_date}
            return data
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_port_ret_resolve_date_range')
    
    def get_port_ret_resolve(self, share_list , begin_date , end_date):
        try:
            basic = BasicDataApi()
            all_ret_dict = []
            for i in range(len(list(share_list.keys()))):
                this_date = list(share_list.keys())[i]
                if i+1 < len(list(share_list.keys())) :
                    next_date = list(share_list.keys())[i+1]
                else :
                    next_date = datetime.date.today()

                if next_date <= begin_date: 
                    continue
                elif next_date <= end_date: 
                    cal_end = next_date
                elif next_date > end_date: 
                    cal_end = end_date

                if this_date >= end_date: 
                    break
                elif this_date <= begin_date: 
                    cal_start = begin_date
                elif this_date > begin_date: 
                    cal_start = this_date
                ret_dict = {}
                for fund_id in list(share_list[this_date].keys()):
                    fund_share = share_list[this_date][fund_id]
                    fund_nav = basic.get_fund_nav_with_date(start_date=cal_start,end_date=cal_end,fund_list=[fund_id])
                    fund_nav = fund_nav.pivot_table(index="datetime" , values="adjusted_net_value" , columns="fund_id").dropna()
                    fund_ret = round((fund_nav.iloc[-1].values[0] * fund_share - fund_nav.iloc[0].values[0] * fund_share)*100,2)
                    ret_dict[fund_id] = fund_ret
                all_ret_dict.append(ret_dict)
            return all_ret_dict
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_port_ret_resolve')

    def get_port_fund_weight(self, fund_wgt, fund_type):
        try:
            fund_info = BasicDataApi().get_fund_info(fund_wgt.columns.tolist())
            name_dic = fund_info.set_index('fund_id').to_dict()['desc_name']
            fund_wgt = fund_wgt.rename(columns=name_dic)
            result = {}
            for types, fund_list in fund_type.items():
                result[types] = [name_dic[_] for _ in fund_list]
            data = {
                '权重':fund_wgt,
                '类型':result,
            }
            return data
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_port_fund_weight')


import matplotlib as mpl
import pylab as pl
import datetime
import matplotlib.pyplot as plt
import pandas as pd
from ...fund.engine.tuner import ScoreTuner
from ...fund.engine.backtest import *
from ...resource.data_store import DataStore
from .score import ScoreFunc
from ...util.calculator import Calculator
from .manager_fund import FundDataManager
from ..struct import AssetTrade, FundTrade, FundScoreParam, TAAParam, TaaTunerParam, AssetTradeParam, AssetWeight, AssetTimeSpan

class FundScoreTool:

    '''
    收益能力因子:  alpha ret_over_period treynor m_square annual_ret time_ret alpha_w jensen_alpha_m treynor_ratio_m
    均衡能力因子:  info_ratio sharpe calmar_ratio_m information_ratio_m sharpe_ratio_m
    风险能力因子:  beta down_risk mdd var annual_vol beta_w beta_m
    跟踪能力因子:  track_err track_err_w
    费用能力因子:  fee_rate
    '''

    INDICATOR_SIGN={
        'alpha':1,
        'beta':-1,
        'fee_rate':-1,
        'track_err':-1,
        'down_risk':-1,
        'info_ratio':1,
        'ret_over_period':1,
        'treynor':1,
        'mdd':-1,
        'm_square':1,
        'var':-1,
        'r_square':0, #解释 beta alpha 有效性
        'sharpe':1,
        'annual_ret':1,
        'annual_vol':-1,
        'time_ret':1,
        'alpha_w':1,
        'beta_w':-1,
        'track_err_w':-1,
        'beta_m':-1,
        'calmar_ratio_m':1,
        'information_ratio_m':1,
        'jensen_alpha_m':1,
        'sharpe_ratio_m':1,
        'treynor_ratio_m':1,
    }

    def __init__(self):
        self.index_list = list(AssetTimeSpan().__dict__.keys())

    def init(self):
        self.dm = DataStore.load_dm()
        self.desc_dic = self.dm.dts.fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']
        self.calculate_std()

    def recent_day(self, dt, duration):
        dts = self.dm.dts.trading_days.datetime
        dts = dts[dts > dt].values
        return dts[0], dts[duration]

    def calculate_std(self):
        dts = self.dm.dts.trading_days.datetime
        dts = dts[dts > datetime.date(2019,1,1)] #todo
        res = []
        for index_id in self.index_list:
            fund_indicator_index = self.dm.dts.fund_indicator[self.dm.dts.fund_indicator.index_id == index_id]
            for dt in dts:
                std_dic = fund_indicator_index[fund_indicator_index.datetime == dt].std(ddof=1).to_dict()
                std_dic['datetime'] = dt
                std_dic['index_id'] = index_id
                res.append(std_dic)
        
        self.indicator_daily_std = pd.DataFrame(res)
        self.indicator_std_mean = self.indicator_daily_std.groupby('index_id').mean()
        self.indicator_list = [i for i in self.indicator_daily_std.columns.tolist() if i not in ['year_lenght','index_id','fund_id']]


    def make_score_equation(self, index_id:str='hs300', score_weight:ScoreFunc=ScoreFunc()):
        score_dict = {}
        for indicator, w in score_weight.__dict__.items():
            if w != 0:
                if indicator in ['beta', 'beta_w', 'beta_m']:
                    score_w = self.INDICATOR_SIGN[indicator] * w
                else:
                    score_w = self.INDICATOR_SIGN[indicator] * w / getattr(self.indicator_std_mean.loc[index_id], indicator)
                score_i = {indicator: score_w}
                score_dict.update(score_i)
        return ScoreFunc(**score_dict)

    def fund_score_tool(self, begin_date:datetime.date=datetime.date(2015,1,1),
                              duration:int=242,
                              fund_num_list:list=[1,2,5,10],
                              score_func:ScoreFunc=ScoreFunc(alpha=0.9,beta=-0.1),
                              is_divide_by_std:bool=True,
                              index_id:str='hs300'):

        begin_date, end_date = self.recent_day(begin_date, duration)
        if is_divide_by_std:
            score_func = self.make_score_equation(index_id, score_func)
        index_price = self.dm.dts.index_price.loc[begin_date:end_date,[index_id]].copy()
        normalized_indicator = self.dm.dts.index_fund_indicator_pack.loc[index_id].loc[begin_date]
        if index_id == 'national_debt':
            time_span = AssetTimeSpan().__dict__[index_id]
            normalized_indicator = normalized_indicator[normalized_indicator.year_length > time_span]    
        score_raw = normalized_indicator.apply(lambda x: score_func.get(x), axis=1)
        top_fund_list = pd.DataFrame(score_raw).sort_values(0,ascending=False).index.tolist()
        indicators = [indicator for indicator, w in score_func.__dict__.items() if w != 0]
        fund_res = []
        for select_num in fund_num_list:
            select_funds = top_fund_list[:select_num] if select_num < len(top_fund_list) else top_fund_list
            col = f'top_{select_num}'
            fund_nav_i = self.dm.dts.fund_nav[select_funds].loc[begin_date:end_date].copy()
            fund_nav_i = fund_nav_i/fund_nav_i.iloc[0]
            fund_nav_i.loc[:,col] = fund_nav_i.mean(axis=1)
            fund_res.append(fund_nav_i[[col]].copy())
        fund_df = pd.concat(fund_res,axis=1,sort=False)
        index_price = index_price / index_price.iloc[0]
        fund_df = fund_df.join(index_price).fillna(method='ffill')
        fund_df.plot.line(figsize=(16,8))
        plt.title('top fund compare', fontsize=20)
        plt.legend(loc='upper right',fontsize=18)
        plt.grid()
        plt.show()
        result_df = normalized_indicator[indicators].loc[top_fund_list[:max(fund_num_list)]]
        result_df.loc[:,'desc_name'] = result_df.index.map(lambda x : self.desc_dic[x])
        result_df.loc[:,'score'] = result_df.index.map(lambda x: score_raw[x])
        result_df = result_df.reset_index()
        result_df.loc[:,'rank'] = result_df.index + 1
        annual_ret = []
        annual_vol = []
        sharpe = []
        mdd = []
        for fund_id in result_df.fund_id:
            res = Calculator.get_stat_result_from_df(df=fund_nav_i[[fund_id]].reset_index(), date_column='datetime', value_column=fund_id)
            annual_ret.append(res.annualized_ret)
            annual_vol.append(res.annualized_vol)
            mdd.append(res.mdd)
            sharpe.append(res.sharpe)
        result_df['mdd'] = mdd
        result_df['annual_ret'] = annual_ret
        result_df['annual_vol'] = annual_vol
        result_df['sharpe'] = sharpe
        res = []
        for group_id in fund_df:
            _res = Calculator.get_stat_result_from_df(df=fund_df[[group_id]].reset_index(), date_column='datetime', value_column=group_id)
            dic = {
                'group':group_id,
                'annual_ret':_res.annualized_ret,
                'annual_vol':_res.annualized_vol,
                'sharpe':_res.sharpe,
                'mdd':_res.mdd,
            }
            res.append(dic)
        col = ['fund_id','desc_name'] + indicators + ['score','annual_ret','mdd','annual_vol','sharpe']
        result_df = result_df.set_index('rank')[col]
        group_result = pd.DataFrame(res)[['group','annual_ret','annual_vol','mdd','sharpe']]
        return group_result, result_df

class SingleIndexFundScore:

    def __init__(self):
        pass

    def backtest(self, m, index_id, max_fund_number, fund_trade_selection, begin_date, score_func):
        saa = AssetWeight(**{index_id:1})
        t = FundTrader(AssetTradeParam(), FundTradeParam(JudgeIndexDiff=1,JudgeFundSelection=fund_trade_selection))
        fa_param = FAParam(MaxFundNumUnderAsset=max_fund_number)
        funcs = {index_id:score_func }
        begin_date = begin_date

        self.bk = FundBacktestEngine(data_manager=m, trader=t, fa_params=fa_param, fund_score_funcs=funcs)
        self.bk.init()
        self.bk.run(saa=saa, start_date=begin_date)
        self.bk_result = self.bk.get_fund_result()

    def get_result(self):
        return pd.DataFrame([self.bk_result])[['annual_ret','annual_vol','sharpe','mdd','mdd_d1','mdd_d2','turnover_rate_yearly_avg','rebalance_times']]

    def plot(self):
        self.bk._report_helper.singel_index_plot()

    def get_rebalance_detail(self):
        return self.bk._report_helper.get_rebalance_detail()

    def get_trade_history(self):
        return self.bk._report_helper.get_fund_trade()



    


from ....util.config import SurfingConfigurator

# key space
cas_default_key_space = SurfingConfigurator().get_cassandra_setting().default_key_space
CAS_KEY_SPACE = 'surfing' if not cas_default_key_space else cas_default_key_space



import pandas as pd
import numpy as np
import datetime
import traceback
import time
from typing import List, Dict, Optional
from functools import partial
from multiprocessing import Pool
from scipy.optimize import Bounds, minimize
import statsmodels.api as sm
from statsmodels import regression
from statsmodels.tsa.ar_model import AutoReg
from pandas.tseries.offsets import DateOffset
from ....util.calculator import Calculator
from ...api.basic import BasicDataApi
from ...api.derived import DerivedDataApi
from ...wrapper.mysql import BasicDatabaseConnector, DerivedDatabaseConnector
from ...view.basic_models import FundNav, Fund_size_and_hold_rate, FundInfo, FundStatusLatest, FundOpenInfo
from ...view.derived_models import FundManagerInfo, FundManagerIndex, FundManagerScore
from .derived_data_helper import DerivedDataHelper, normalize, score_rescale
from ...manager.fund_info_filter import EHR_PROBLEM_EKYWORDS


class ManagerProcessorDev:

    #INDEX_LIST = {'csi_boodfund': 'H11023.CSI', 'csi_f_hybrid': 'H11022.CSI', 'csi_f_qdii': 'H11026.CSI', 'csi_stockfund': 'H11021.CSI', 'mmf': 'H11025.CSI','hs300':'hs300'}
    #REPLACE_DICT = {'stock': 'H11021.CSI', 'bond': 'H11023.CSI', 'mmf': 'H11025.CSI', 'QDII': 'H11026.CSI', 'index': 'hs300'}
    #DATE_DICT = {'stock': datetime.date(2007,1,1), 'bond': datetime.date(2003,1,1), 'mmf': datetime.date(2004,1,7), 'QDII': datetime.date(2008,1,1), 'index': datetime.date(2007,1,1)}

    FUND_CLASSIFIER = {
        'stock': ['普通股票型基金', '偏股混合型基金',  '被动指数型基金', '增强指数型基金', '平衡混合型基金', '灵活配置型基金', '股票多空'],
        'bond': ['中长期纯债型基金', '短期纯债型基金', '偏债混合型基金', '增强指数型债券基金', '被动指数型债券基金','混合债券型二级基金', '混合债券型一级基金'], 
        'QDII': ['国际(QDII)股票型基金', '国际(QDII)债券型基金', '国际(QDII)另类投资基金', '国际(QDII)混合型基金', ],
        'mmf': ['货币市场型基金'],
        'index': ['被动指数型基金', '增强指数型基金', '增强指数型债券基金', '被动指数型债券基金', '商品型基金', 'REITs'],
    }
    REPLACE_DICT = {'stock': 'csi_stockfund',
                     'bond': 'national_debt', 
                     'mmf':  'mmf', 
                     'QDII': 'csi_f_qdii', 
                     'index': 'hs300'}
    FILTER_NAV_RATIO = {
        'stock' : 0.1,
        'QDII'  : 0.1,
        'index' : 0.1,
        'bond'  : 0.07,
        'mmf'   : 0.01,
    }

    INDEX_DICT = { v : k for k, v in REPLACE_DICT.items()}
    PUNISH_RATIO = 0.8
    HARD_PUNISH_RATIO = 0.6
    RISK_FEE_RATE = 0.011
    TRADE_DAY_PER_YEAR = 242
    RISK_FEE_RATE_PER_DAY = RISK_FEE_RATE / TRADE_DAY_PER_YEAR
    TRADING_DAYS_PER_MONTH = 20
    # 管理不足 1 年 0.6； 大于1年小于三年 0.8， 三年以上 1

    def __init__(self, data_helper: DerivedDataHelper):
        self._data_helper: DerivedDataHelper = data_helper
        self._basic_api = BasicDataApi()
        self.mng_index_list: Dict[str, pd.DataFrame] = {}  # 基金经理指数
        self.mng_best_fund: Dict[str, pd.DataFrame] = {}  # 基金经理代表作 
        self.mng_indicator_list: Dict[str, pd.DataFrame] = {}  # 基金经理评分

    def init(self, start_date=str, end_date=str, print_time=False, is_mng_index=False, is_mng_indicator=False):
        self.start_date = start_date if isinstance(start_date, datetime.date) else pd.to_datetime(start_date, infer_datetime_format=True).date()
        self.end_date = end_date if isinstance(end_date, datetime.date) else pd.to_datetime(end_date, infer_datetime_format=True).date()

        with BasicDatabaseConnector().managed_session() as quant_session:
            # 交易日序列
            self.trading_days: pd.DataFrame = self._basic_api.get_trading_day_list().set_index('datetime')
            self.trading_days = self.trading_days.loc[: self.end_date]
            self.trading_days_list = self.trading_days.index
            self.trim_days = self.trading_days[(self.trading_days.index >= self.start_date) & (self.trading_days.index <= self.end_date)].index

            # 指数价格
            self.index_price = self._basic_api.get_index_price(index_list=list(self.REPLACE_DICT.values()))
            self.index_price['index_id'] = self.index_price['index_id'].map(self.INDEX_DICT)
            self.index_price = self.index_price.pivot(index='datetime', columns='index_id', values='close').reindex(self.trim_days).ffill().dropna(how='all', axis=1)
            self.index_ret = np.log(self.index_price).diff()
            
            # 基金信息
            fund_info_query = quant_session.query(FundInfo)
            self.fund_info: pd.DataFrame = pd.read_sql(fund_info_query.statement, fund_info_query.session.bind)
        
            ## 计算基金经理指数时只去掉分级基金子基金， 不能回测交易的基金依然影响基金经理能力
            self.fund_info = self.fund_info[(self.fund_info.structure_type <= 1)
                                          & (~self.fund_info.wind_class_2.isnull())]
            # 基金规模 报告期数据填充交易日 然后再只取交易日，保证和基金收益序列一致
            fund_size_query = quant_session.query(
                Fund_size_and_hold_rate.fund_id,
                Fund_size_and_hold_rate.datetime,
                Fund_size_and_hold_rate.size
            )
            self.fund_size = pd.read_sql(fund_size_query.statement, fund_size_query.session.bind)
            self.fund_size = self.fund_size.pivot_table(index='datetime', columns='fund_id', values='size')
            self.fund_size = self.fund_size.reindex(self.fund_size.index.union(self.trim_days)).dropna(axis=1, how='all')
            self.fund_size = self.fund_size.ffill()

            if is_mng_index:
                # 基金净值 和 日收益
                fund_nav_query = quant_session.query(
                    FundNav.fund_id,
                    FundNav.adjusted_net_value,
                    FundNav.datetime
                ).filter(
                    FundNav.datetime >= self.start_date,
                    FundNav.datetime <= self.end_date,
                    FundNav.fund_id.in_(self.fund_info.fund_id)
                )
                self.fund_nav = pd.read_sql(fund_nav_query.statement, fund_nav_query.session.bind)
                self.fund_nav = self.fund_nav.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value').reindex(self.trim_days).dropna(axis=1, how='all')
                self.fund_nav = self.fund_nav.ffill() + self.fund_nav.bfill() - self.fund_nav.bfill()
                self.fund_ret = np.log(self.fund_nav).diff() 
                
                # 按照不同基金大类类别做净值过滤， 参考FILTER_NAV_RATIO
                _res = []
                _fund_ret_columns = self.fund_ret.columns.tolist()
                for fund_type, filter_ratio in self.FILTER_NAV_RATIO.items():
                    _info_fund_list = self.fund_info[self.fund_info['wind_class_2'].isin(self.FUND_CLASSIFIER[fund_type])].fund_id.tolist()
                    _fund_list = list(set(_info_fund_list).intersection(_fund_ret_columns))
                    fund_ret = self.fund_ret[_fund_list].copy()
                    fund_ret[fund_ret > filter_ratio] = 0
                    _res.append(fund_ret)
                self.fund_ret = pd.concat(_res,axis=1)
                self.fund_ret = self.fund_ret.loc[:,~self.fund_ret.columns.duplicated()].copy()
                self.fund_size = self.fund_size.reindex(self.fund_ret.index)

            elif is_mng_indicator:
                self.fund_manager_index: pd.DataFrame = DerivedDataApi().get_fund_manager_index().drop(columns='_update_time')
                self.fund_manager_index.groupby(by='fund_type', sort=False).apply(self._preprocess_manager_index_df)
                self.fund_size = self.fund_size.reindex(self.trim_days)

        with DerivedDatabaseConnector().managed_session() as quant_session:
            # 基金经理信息
            mng_query = quant_session.query(
                FundManagerInfo.mng_id,
                FundManagerInfo.start_date,
                FundManagerInfo.fund_id,
                FundManagerInfo.mng_name,
                FundManagerInfo.end_date)
            self.fund_manager_info: pd.DataFrame = pd.read_sql(mng_query.statement, mng_query.session.bind)
            self.fund_manager_info = self.fund_manager_info.rename(columns={'mng_id': 'manager_id', 'mng_name': 'name', 'start_date': 'start', 'end_date': 'end'})
            self.fund_manager_info = pd.merge(self.fund_manager_info, self.fund_info[['fund_id','company_id','wind_class_2']], how='left', on='fund_id' )

        self.manager_name_dict = self.fund_manager_info.set_index('manager_id').to_dict()['name'] 
        self.fund_manager_info = self.fund_manager_info.set_index('fund_id')
        # 每个基金经理人数   如果多人给惩罚
        fund_list = self.fund_manager_info.index.unique().tolist()
        self.fund_manager_num_df = self.process_manager_num(fund_list, self.fund_ret.index)
        
        # 每个基金的规模  小于1e 给惩罚
        self.fund_manager_size_df = self.fund_size.copy()
        self.fund_manager_size_df[self.fund_manager_size_df < 1e8] = 0.6
        self.fund_manager_size_df[(self.fund_manager_size_df >= 1e8) & (self.fund_manager_size_df < 2e9)] = 0.7
        self.fund_manager_size_df[(self.fund_manager_size_df >= 2e9) & (self.fund_manager_size_df < 5e9)] = 0.8
        self.fund_manager_size_df[(self.fund_manager_size_df >= 5e9) & (self.fund_manager_size_df < 1e10)] = 0.9
        self.fund_manager_size_df[self.fund_manager_size_df >= 1e10] = 1
        self.fund_manager_size_df = self.fund_manager_size_df.fillna(self.PUNISH_RATIO)

        # 基金经理数据 根据被动主动分类
        pass_fund_id_list = self.fund_info[self.fund_info.wind_class_2.str.contains('被动')].fund_id.tolist()
        pass_fund_id_list = [i for i in pass_fund_id_list if i in self.fund_ret.columns] 
        acti_fund_id_list = [i for i in self.fund_ret.columns if i not in pass_fund_id_list] 
        _df_acti = pd.DataFrame(1, columns=acti_fund_id_list, index=self.fund_ret.index)
        _df_pass = pd.DataFrame(self.PUNISH_RATIO, columns=pass_fund_id_list, index=self.fund_ret.index)
        self.fund_manager_stype_df = pd.concat([_df_acti, _df_pass], axis=1)
        self.period = ['1y', '3y', '5y', 'history']

    def get_resample_ret(self, df, rule='1M', min_count=15):

        return df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).sum(min_count=min_count)

    def get_begin_date(self, period: str):
        if period == '1y':
            return self.trading_days.tail(self.TRADE_DAY_PER_YEAR * 1).index[0]

        if period == '3y':
             return self.trading_days.tail(self.TRADE_DAY_PER_YEAR * 3).index[0]

        if period == '5y':
             return self.trading_days.tail(self.TRADE_DAY_PER_YEAR * 5).index[0]

        if period == 'history':
            return  self.trading_days.index[0]

    def calc_stutzer(self, ex_return: np.ndarray):
        information_statistic = lambda theta: np.log(np.mean(np.exp(theta[0] * ex_return)))
        theta0 = [-1.]
        bounds = Bounds((-np.inf), (0))
        result = minimize(information_statistic, theta0, method='SLSQP', bounds=bounds, tol=1e-16)
        if result.success:
            information_statistic = -result.fun
            if information_statistic <= 0:
                return 0
            stutzer_index = np.abs(np.mean(ex_return)) / np.mean(ex_return) * np.sqrt(2 * information_statistic)
            return stutzer_index
        else:
            return np.nan
    
    def calc_cl_alpha_beta(self, total: np.ndarray):
        if total.shape[0] <= 1:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 0][X[:, 0] < 0] = 0
        X[:, 1][X[:, 1] > 0] = 0
        if np.count_nonzero(X[:, 1]) == 0:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[0] - est2.params[1],
                'alpha':est2.params[-1],
                }

    def calc_continue_regress_v(self, df, period='monthly'):
        if period == 'monthly':
            # 半年周期
            window = 6
            yearly_multiplier = 2
            period_num = 12
        df.loc[:,'annual_ret'] = df.mng_ret.rolling(window=window).apply(lambda x : x.sum() * yearly_multiplier - self.RISK_FEE_RATE)
        df.loc[:,'annual_vol'] = df.mng_ret.rolling(window=window).apply(lambda x : x.std(ddof=1) * np.sqrt(period_num))
        df.loc[:,'sharpe'] = df.loc[:,'annual_ret'] / df.loc[:,'annual_vol']
        df = df.replace(np.Inf, 0).replace(-np.Inf, 0).dropna() # 空值sharpe赋值0
        if df.shape[0] < 6:
            return np.nan
        mod = AutoReg(endog=df['sharpe'].dropna().values, lags=1)
        res = mod.fit()
        continue_regress_v = res.params[0]
        return continue_regress_v

    def lambda_filter_date_range(self, single_fund_info: pd.DataFrame, date_index: pd.Series) -> pd.Series:
        result: List[pd.Series] = []
        for row in single_fund_info.itertuples(index=False):
            result.append(pd.Series((date_index >= row.start) & (date_index <= row.end), index=date_index))
        return pd.concat(result, axis=1).sum(axis=1)

    def process_fund_ret(self, this_manager_fund_ids, this_manager_info, date_index):
        res = []
        for fund_id in this_manager_fund_ids:
            single_fund_info = this_manager_info.loc[[fund_id]]
            res_i =  self.lambda_filter_date_range(single_fund_info, date_index)
            res_i.name = fund_id
            res.append(res_i)
        return pd.concat(res, axis=1)

    def process_manager_num(self, fund_id_list, date_index):
        res = []
        for fund_id in fund_id_list:
            fund_managers_info = self.fund_manager_info.loc[[fund_id]]
            fund_manager_list = fund_managers_info.manager_id.unique()
            _res = []
            for _mng_id in fund_manager_list:
                fund_manager_i = fund_managers_info[fund_managers_info['manager_id'] == _mng_id]
                single_fund_res = self.lambda_filter_date_range(fund_manager_i, date_index)
                single_fund_res = pd.DataFrame(single_fund_res, columns=[_mng_id])
                _res.append(single_fund_res)
            manager_in_same_time = pd.concat(_res,axis=1)
            fund_manger_same_time = pd.DataFrame(manager_in_same_time.sum(axis=1), columns=[fund_id])            
            res.append(fund_manger_same_time)
        fund_manger_same_time = pd.concat(res, axis=1)
        fund_manger_same_time[fund_manger_same_time > 1] = self.PUNISH_RATIO
        return fund_manger_same_time

    def process_manager_experience(self, fund_id_list, this_manager_info, date_index):
        res = []
        for fund_id in fund_id_list:
            single_fund_info = this_manager_info.loc[[fund_id]]
            work_day_df_i = self.lambda_filter_date_range(single_fund_info, date_index)
            work_day_df_i.name = fund_id
            res.append(work_day_df_i)
        work_day_df = pd.DataFrame(res).T
        work_day_df = work_day_df.cumsum()
        low_cons = work_day_df < self.TRADE_DAY_PER_YEAR
        mid_cons = (work_day_df >= self.TRADE_DAY_PER_YEAR) & (work_day_df < 3 * self.TRADE_DAY_PER_YEAR)
        hig_cons = work_day_df >= 3 * self.TRADE_DAY_PER_YEAR
        work_day_df[low_cons] = self.HARD_PUNISH_RATIO
        work_day_df[mid_cons] = self.PUNISH_RATIO
        work_day_df[hig_cons] = 1
        return work_day_df

    def calc_index(self):
        # 遍历每个基金经理
        #_tm_part2_start = time.time()
        for fund_type in self.FUND_CLASSIFIER:
            self.mng_index_list[fund_type] = pd.DataFrame(index=self.fund_ret.index)
            self.mng_best_fund[fund_type] = pd.DataFrame(index=self.fund_ret.index)

        fund_manager_list = list(self.manager_name_dict.keys())
        t0 = time.time()
        for m_id in fund_manager_list:
            _t0 = time.time()
            for fund_type in self.FUND_CLASSIFIER:
                idx = self.index_calculation(m_id, fund_type)
                if idx is not None:
                    self.mng_index_list[fund_type].loc[:, m_id] = idx # 该资产类别下记录计算结果

            _idx = fund_manager_list.index(m_id)
            _t1 = time.time()
            #print(f'm_id {m_id} {_idx} {len(fund_manager_list)} cost time {_t1 - _t0}')
        t1 = time.time()
        print(f'total cost {t1 - t0}')
        for fund_type in self.FUND_CLASSIFIER:
            self.mng_index_list[fund_type] = self.mng_index_list[fund_type].replace(0, np.nan).dropna(how='all')
        # _tm_part2_end = time.time()
        # print(f' part2 cost time {_tm_part2_end - _tm_part2_start}')

    def index_calculation(self, m_id: str, fund_type: str):
        
        wind_class_2_list = self.FUND_CLASSIFIER[fund_type]
        # 该基金经理管理基金信息
        this_manager_info = self.fund_manager_info[(self.fund_manager_info['wind_class_2'].isin(wind_class_2_list))
                            & (self.fund_manager_info['manager_id'] == m_id)]
        if this_manager_info.shape[0] == 0:
            return 
        index_start_date = this_manager_info.start.min()
        index_end_date = this_manager_info.end.max()
        
        # 该基金经理管理的同类基金列表
        this_manager_fund_ids = this_manager_info.index.unique().tolist()
        _fund_id_list1 = self.fund_ret.columns.tolist()
        _fund_id_list2 = self.fund_manager_size_df.columns.tolist()
        this_manager_fund_ids = list(set(this_manager_fund_ids).intersection(_fund_id_list1).intersection(_fund_id_list2))
        if len(this_manager_fund_ids) == 0:
            return
        fund_ret = self.fund_ret.loc[index_start_date:index_end_date, this_manager_fund_ids]

        # 非管理时间，基金收益赋值0
        manager_status = self.process_fund_ret(this_manager_fund_ids, this_manager_info, fund_ret.index)
        fund_ret = fund_ret * manager_status

        # 基金经理管理的基金 人数状态 单人为 1 ，多人为0.8
        manager_num = self.fund_manager_num_df.loc[index_start_date:index_end_date, this_manager_fund_ids]

        # 基金经理管理的基金 规模状态  > 100e : 1, > 50e : 0.9, > 20e : 0.8, > 1e : 0.7, < 1e : 0.6
        manager_size = self.fund_manager_size_df.loc[index_start_date:index_end_date, this_manager_fund_ids]

        # 基金经理管理的基金 负责年限 大于3 1， 大于1 0.8， 小于1 0.6
        manager_year = self.process_manager_experience(this_manager_fund_ids, this_manager_info, fund_ret.index)

        # 基金经理管理的基金 主动 1 , 被动 0.8
        manager_style = self.fund_manager_stype_df.loc[index_start_date:index_end_date, this_manager_fund_ids]

        # 权重累乘
        fund_weight = manager_num * manager_size * manager_year * manager_style * manager_status

        # 权重和归1
        fund_weight['sum'] = fund_weight.sum(axis=1).T
        fund_weight = fund_weight[this_manager_fund_ids].divide(fund_weight['sum'], axis=0)

        # 基金经理 收益
        total_ret = (fund_ret * fund_weight).sum(axis=1)
        total_ret.name = m_id
        total_ret = pd.DataFrame(total_ret)
        if total_ret.shape[0] == 0:
            return 
        # 基金经理 收益 用指数填充待业期
        index_date_list = manager_status.index[manager_status.sum(axis=1) < 1]
        total_ret.loc[index_date_list, m_id] = self.index_ret.loc[index_date_list, fund_type]

        # 收益初值赋0
        last_trading_day = self.trading_days_list[self.trading_days_list < total_ret.index.values[0]].values[-1]
        total_ret.loc[last_trading_day] = 0
        total_ret = total_ret.sort_index()

        # 做指数
        fund_manager_index = np.exp(total_ret.cumsum())

        # 基金经理代表作
        manager_best_fund = pd.DataFrame((fund_weight + manager_status.cumsum()/1e8).idxmax(axis=1), columns=[m_id])
        self.mng_best_fund[fund_type][m_id] = manager_best_fund
        return fund_manager_index

    def calcualtor(self, df, period='monthly'):
        if period == 'monthly':
            yearly_multiplier = 12
            period_days = self.TRADING_DAYS_PER_MONTH
        x = df.benchmark_ret.values
        y = df.mng_ret.values
        x = sm.add_constant(x)
        model = regression.linear_model.OLS(y,x).fit()
        x = x[:,1]
        alpha = model.params[0] * yearly_multiplier
        beta = model.params[1]
        track_err = (y - x).std(ddof=1) * np.sqrt(yearly_multiplier) 
        downside_std = y[y < (self.RISK_FEE_RATE_PER_DAY * period_days)].std(ddof=1)
        info_ratio = alpha / track_err
        treynor = (np.nanmean(y) - self.RISK_FEE_RATE_PER_DAY * period_days) / beta if beta != 0 else np.nan
        mdd_part =  df.mng_index[:] / df.mng_index[:].rolling(window=df.shape[0], min_periods=1).max()
        mdd = 1 - mdd_part.min()
        if np.isnan(mdd):
            mdd_len = np.nan
        else:
            mdd_date2 = mdd_part.idxmin()
            mdd_date1 = df.mng_index[:mdd_date2].idxmax()
            mdd_len = (mdd_date2 - mdd_date1).days

        annual_ret = df.mng_ret.sum() * (yearly_multiplier / df.shape[0]) - self.RISK_FEE_RATE
        annual_vol = df.mng_ret.std(ddof=1) * np.sqrt(yearly_multiplier)
        sharpe = annual_ret / annual_vol
        calma_ratio = annual_ret / mdd if mdd != 0 else np.nan
        var = np.quantile(df.mng_ret, 0.05)
        cvar = df.mng_ret[df.mng_ret <= var].mean()
        stutzer = self.calc_stutzer(df.mng_ret - self.RISK_FEE_RATE_PER_DAY * period_days) * np.sqrt(yearly_multiplier)
        cl_res = self.calc_cl_alpha_beta(df[['mng_ret','benchmark_ret']].to_numpy())
        stock_cl_alpha = cl_res['alpha']
        stock_cl_beta = cl_res['beta']
        return {
            'alpha':alpha,
            'beta':beta,
            'track_err':track_err,
            'downside_std':downside_std,
            'info_ratio':info_ratio,
            'treynor':treynor,
            'mdd':mdd,
            'mdd_len':mdd_len,
            'annual_vol':annual_vol,
            'annual_ret':annual_ret,
            'sharpe':sharpe,
            'calma_ratio':calma_ratio,
            'var':var,
            'cvar':cvar,
            'stutzer':stutzer,
            'stock_cl_alpha':stock_cl_alpha,
            'stock_cl_beta':stock_cl_beta,
        }

    def loop_item(self, mng_id, fund_type):
        _t0 = time.time()
        idx = self.type_manager_list.index(mng_id)
        wind_class_2_list = self.FUND_CLASSIFIER[fund_type]
        this_manager_whole_info = self.fund_manager_info[(self.fund_manager_info['manager_id'] == mng_id)]
        this_manager_info = self.fund_manager_info[(self.fund_manager_info['wind_class_2'].isin(wind_class_2_list))
                            & (self.fund_manager_info['manager_id'] == mng_id)]
        year_df = self.lambda_filter_date_range(this_manager_info, self.fund_ret.index)

        _fund_list = self.fund_size.columns.tolist()
        still_work_fund_ids = this_manager_info[(this_manager_info.end > self.end_date)].index.unique().tolist()
        still_work_fund_ids = [ i for i in still_work_fund_ids if i in _fund_list ]
        mng_ret = self.monthly_ret[[mng_id]].rename(columns={mng_id:'mng_ret'})
        mng_index = self.monthly_nav[[mng_id]].rename(columns={mng_id:'mng_index'})
        mng_data = pd.concat([mng_ret, mng_index, self.benchmark_monthly_ret], axis=1).dropna()
        mng_result = {
            'mng_id': mng_id,
            'datetime': self.end_date
        }
        for period in self.period:
            begin_date = self.get_begin_date(period)
            mng_data_period =  mng_data.loc[begin_date:].copy().dropna()
            if all(mng_data_period.mng_ret == 0) or mng_data_period.shape[0] < 12:
                continue
            period_result = self.calcualtor(mng_data_period, 'monthly')
            period_result = { f'{k}~{period}_m' : v for k, v in period_result.items() }
            mng_result.update(period_result)
        _weekly_ret = self.weekly_ret[mng_id]
        _weekly_ret.name = 'mng_ret'
        _weekly_bench_ret = self.benchmark_weekly_ret
        weekly_data = pd.concat([_weekly_ret,_weekly_bench_ret], axis=1).dropna()
        cl_res = self.calc_cl_alpha_beta(weekly_data[['mng_ret','benchmark_ret']].to_numpy())
        stock_cl_alpha = cl_res['alpha']
        stock_cl_beta = cl_res['beta']
        mng_result.update({'stock_cl_alpha~history_w':stock_cl_alpha})
        mng_result.update({'stock_cl_beta~history_w':stock_cl_beta})  
        mng_result.update({'fund_type':fund_type})  
        mng_result.update({'fund_type_trading_days': int((year_df / year_df).sum())})  
        mng_result.update({'trading_days': (self.end_date - this_manager_whole_info.start.min()).days})  
        mng_result.update({'scale': self.fund_size.loc[self.end_date, still_work_fund_ids].sum()})
        mng_result.update({'total_ret': self.fund_type_df[mng_id].values[-1]})
        _mng_index = self.fund_type_df[mng_id]
        mng_result.update({'mdd~history_daily': 1 - (_mng_index / _mng_index.cummax()).min()})
        _t1 = time.time()
        #print(f'{mng_id} {idx} / {len(self.type_manager_list)} cost time {_t1 - _t0}')
        return mng_result

    def calc_mng_indicator(self):
        self.result = []
        for fund_type in self.FUND_CLASSIFIER:            
            self.fund_type_df = self.mng_index_list[fund_type]
            still_work_mng_list = self.fund_type_df.iloc[-1].dropna().index.tolist()
            daily_ret = np.log(self.fund_type_df[still_work_mng_list]).diff()
            self.weekly_ret = self.get_resample_ret(df=daily_ret, rule='1W', min_count=3).dropna(axis=0, how='all')
            self.weekly_ret.index = self.weekly_ret.index.date
            self.monthly_ret = self.get_resample_ret(df=daily_ret, rule='1M', min_count=15).dropna(axis=0, how='all')
            self.monthly_ret.index = self.monthly_ret.index.date
            self.type_manager_list = self.monthly_ret.columns.tolist()
            self.monthly_nav = np.exp(self.monthly_ret.cumsum())
            daily_index_ret = np.log(self.index_price).diff()
            self.index_monthly_ret = self.get_resample_ret(df=daily_index_ret, rule='1M', min_count=15).dropna(axis=0, how='all')
            self.index_monthly_ret.index = self.index_monthly_ret.index.date
            self.index_weekly_ret = self.get_resample_ret(df=daily_index_ret, rule='1W', min_count=3).dropna(axis=0, how='all')
            self.index_weekly_ret.index = self.index_weekly_ret.index.date
            self.benchmark_monthly_ret = self.index_monthly_ret[[fund_type]].rename(columns={fund_type:'benchmark_ret'})
            self.benchmark_weekly_ret = self.index_weekly_ret[[fund_type]].rename(columns={fund_type:'benchmark_ret'})
            for mng_id in self.type_manager_list:
                res_i = self.loop_item(mng_id, fund_type)
                self.result.append(res_i)
        self.manager_indicator = pd.DataFrame(self.result)

    def calc_score(self):
        res=[]
        for fund_type in self.FUND_CLASSIFIER:
            df = self.manager_indicator[self.manager_indicator['fund_type'] == fund_type].copy()
            if fund_type == 'stock':
                df.loc[:,'ret_ability'] = score_rescale( 0.4 * score_rescale(df['annual_ret~history_m']) + 0.6 * score_rescale(df['total_ret']))
                df.loc[:,'risk_ability'] = score_rescale(0.3 * score_rescale(-df['mdd~history_daily']) + 0.4 * score_rescale(-df['annual_vol~history_m']) + 0.3 * score_rescale(-df['downside_std~history_m']))
                df.loc[:,'select_time'] =  score_rescale(normalize(df['stock_cl_beta~history_w']))
                df.loc[:,'select_stock'] = score_rescale(normalize(df['stock_cl_alpha~history_w']) )
                df.loc[:,'experience'] = score_rescale( 0.8 *  score_rescale(df['fund_type_trading_days']) + 0.2 * score_rescale(df['scale']))
                df.loc[:,'total_score'] = score_rescale(df.loc[:,'ret_ability'] + df.loc[:,'risk_ability'] + df.loc[:,'select_time'] + df.loc[:,'select_stock'] + 1.5 * df.loc[:,'experience'])
            
            elif fund_type == 'bond':
                df.loc[:,'ret_ability'] = score_rescale( 0.8 * score_rescale(df['annual_ret~history_m']) + 0.2 * score_rescale(df['total_ret']))
                df.loc[:,'risk_ability'] = score_rescale(0.1 * score_rescale(-df['mdd~history_daily']) + 0.4 * score_rescale(-df['annual_vol~history_m']) + 0.5 * score_rescale(-df['downside_std~history_m']))
                df.loc[:,'select_time'] =  score_rescale(normalize(df['stock_cl_beta~history_w']))
                df.loc[:,'select_stock'] = score_rescale(normalize(df['stock_cl_alpha~history_w']))
                df.loc[:,'experience'] = score_rescale( 0.8 *  score_rescale(df['fund_type_trading_days']) + 0.2 * score_rescale(df['scale']))
                df.loc[:,'total_score'] = score_rescale(1.2 * df.loc[:,'ret_ability'] +  df.loc[:,'risk_ability'] + df.loc[:,'select_time'] + df.loc[:,'select_stock'] + 1.5 * df.loc[:,'experience'])
            
            elif fund_type == 'index':
                df.loc[:,'ret_ability'] = score_rescale( 0.4 * score_rescale(df['annual_ret~history_m']) + 0.6 * score_rescale(df['total_ret']))
                df.loc[:,'risk_ability'] = score_rescale(0.3 * score_rescale(-df['mdd~history_daily']) + 0.4 * score_rescale(-df['annual_vol~history_m']) + 0.3 * score_rescale(-df['downside_std~history_m']))
                df.loc[:,'experience'] = score_rescale( 0.4 *  score_rescale(df['fund_type_trading_days']) + 0.6 * score_rescale(df['scale']))
                df.loc[:,'total_score'] = score_rescale(df.loc[:,'ret_ability'] + df.loc[:,'risk_ability'] + 2 * df.loc[:,'experience'] )
            
            elif fund_type == 'QDII':
                df.loc[:,'ret_ability'] = score_rescale( 0.4 * score_rescale(df['annual_ret~history_m']) + 0.6 * score_rescale(df['total_ret']))
                df.loc[:,'risk_ability'] = score_rescale(0.3 * score_rescale(-df['mdd~history_daily']) + 0.4 * score_rescale(-df['annual_vol~history_m']) + 0.3 * score_rescale(-df['downside_std~history_m']))
                df.loc[:,'stable_ability'] = score_rescale(-normalize(df['annual_vol~history_m']))
                df.loc[:,'select_time'] = np.nan
                df.loc[:,'select_stock'] = np.nan
                df.loc[:,'experience'] = score_rescale( 0.8 *  score_rescale(df['fund_type_trading_days']) + 0.2 * score_rescale(df['scale']))
                df.loc[:,'total_score'] = score_rescale(df.loc[:,'ret_ability'] + df.loc[:,'risk_ability'] + df.loc[:,'stable_ability'] + 1.5 * df.loc[:,'experience'])
            
            elif fund_type == 'mmf':
                df.loc[:,'ret_ability'] = score_rescale(normalize(df['annual_ret~1y_m']))
                df.loc[:,'experience'] = score_rescale( 0.8 *  score_rescale(df['fund_type_trading_days']) + 0.2 * score_rescale(df['scale']))
                df.loc[:,'total_score'] = score_rescale(df.loc[:,'ret_ability'] + df.loc[:,'experience'])
            res.append(df)
        self.manager_indicator_score = pd.concat(res)
        _df = self.fund_manager_info.reset_index()
        _df = _df[_df.end == datetime.date(2040,1,1)].dropna(subset=['company_id']).copy()
        _df = _df.rename(columns={'manager_id':'mng_id'})[['mng_id','name','company_id']].drop_duplicates(subset=['mng_id','name']).set_index('mng_id')
        self.manager_indicator_score = self.manager_indicator_score.set_index('mng_id').join(_df).reset_index()
    
    def to_s3(self):
        bucket_name = 'tl-fund-dm'
        _s3_bucket_uri: str = f's3://{bucket_name}/dm'

        name = 'mng_best_fund'
        data = self.mng_best_fund
        res_dict = {}
        for k, df in data.items():
            res_dict[k] = df.to_dict('index')
        df = pd.DataFrame(res_dict)    
        df.to_parquet(f'{_s3_bucket_uri}/{name}.parquet', compression='gzip')

        name = 'mng_index_list'
        data = self.mng_index_list
        res_dict = {}
        for k, df in data.items():
            res_dict[k] = df.to_dict('index')
        df = pd.DataFrame(res_dict)    
        df.to_parquet(f'{_s3_bucket_uri}/{name}.parquet', compression='gzip')

        name = 'mng_indicator_score'
        self.manager_indicator_score.to_parquet(f'{_s3_bucket_uri}/{name}.parquet', compression='gzip')

    def process(self, end_date):
        failed_tasks = []
        try:
            start_date = '20050101'
            self.init(start_date=start_date,end_date=end_date,is_mng_index=True)
            self.calc_index()
            self.calc_mng_indicator()
            self.calc_score()
            self.to_s3()
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('manager_indicator_score')
        return failed_tasks


if __name__ == '__main__':
    mpd = ManagerProcessorDev(DerivedDataHelper())
    mpd.process('20210226')





from typing import Optional, List
import re
import datetime
import os
import traceback
import time

import pandas as pd
import pdfplumber

from ...util.mail_retriever import MailAttachmentRetriever, IMAP_SPType, UID_FILE_NAME
from ...util.wechat_bot import WechatBot
from ..wrapper.mysql import BasicDatabaseConnector
from ..view.basic_models import HedgeFundNAV
from ..api.basic import BasicDataApi


class HedgeFundNAVReader:

    COLUMNS_DICT = {
        '基金代码': 'fund_id',
        '产品代码': 'fund_id',
        '基金名称': 'fund_name',
        '产品名称': 'fund_name',
        '基金份额净值': 'net_asset_value',
        '单位净值': 'net_asset_value',
        '计提前单位净值': 'net_asset_value',
        '实际净值': 'net_asset_value',
        '基金份额累计净值': 'acc_unit_value',
        '累计单位净值': 'acc_unit_value',
        '累计净值': 'acc_unit_value',
        '虚拟后净值': 'v_net_value',
        '虚拟净值': 'v_net_value',
        '计提后单位净值': 'v_net_value',
        '虚拟计提净值': 'v_net_value',
        '虚拟后单位净值': 'v_net_value',
        '复权累计净值': 'adjusted_net_value',
        '复权净值': 'adjusted_net_value',
        '净值(分红再投)': 'adjusted_net_value',
        '日期': 'datetime',
        '净值日期': 'datetime',
        '业务日期': 'datetime',
        '计算日期': 'calc_date',
    }

    def __init__(self, read_dir: str, user_name: str, password: str):
        pd.set_option('display.max_rows', None)
        pd.set_option('display.max_columns', None)

        self._read_dir = read_dir
        assert os.path.isdir(self._read_dir), f'arg dump_dir should be a directory (now){self._read_dir}'

        self._user_name = user_name
        self._password = password
        self._wechat_bot = WechatBot()

    @staticmethod
    def _read_for_HuaChengZhiYuanNo3WithoutV(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path, header=None)
        df.iloc[:, 0] = df.iloc[:, 0].map(lambda x: x.split('：')[0])
        df = df.set_index(df.columns[0])
        pat = r'\d+'
        date = re.findall(pat, df.index.array[2])
        df = df.T.loc[:, ['基金代码', '基金名称', '基金份额净值', '基金份额累计净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT).assign(datetime=pd.to_datetime('-'.join(date), infer_datetime_format=True).date())
        return df

    @staticmethod
    def _read_for_HuaChengZhiYuanNo3(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品代码', '产品名称', '业务日期', '单位净值', '累计单位净值', '虚拟后净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        df['datetime'] = pd.to_datetime(df.datetime.astype(str), infer_datetime_format=True).dt.date
        return df

    @staticmethod
    def _read_for_TianYanGuangQuanWithoutV(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品名称', '产品代码', '净值日期', '单位净值', '累计单位净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        return df

    @staticmethod
    def _read_for_TianYanGuangQuan(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path, skipfooter=1)
        df = df.loc[:, ['基金名称', '基金代码', '净值日期', '计算日期', '单位净值', '累计单位净值', '虚拟净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        return df

    @staticmethod
    def _read_for_AnXianHuaMuChangSheng(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品名称', '产品代码', '日期', '单位净值', '累计单位净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        df = df.replace({'S3ZS26': 'SJE335'})
        df['datetime'] = df.datetime.transform(lambda x: pd.to_datetime(str(x), infer_datetime_format=True).date())
        return df

    @staticmethod
    def _read_for_WuZhiCTA1(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品名称', '产品代码', '净值日期', '计提前单位净值', '累计单位净值', '计提后单位净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        return df

    @staticmethod
    def _read_for_WuZhiCTA1WithoutV(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品名称', '产品代码', '日期', '单位净值', '累计单位净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        return df

    @staticmethod
    def _read_for_XinPu10(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品名称', '产品代码', '业务日期', '单位净值', '累计单位净值', '虚拟后净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        df['datetime'] = df.datetime.transform(lambda x: pd.to_datetime(str(x), infer_datetime_format=True).date())
        return df

    @staticmethod
    def _read_for_TaiChuangXiShuoCTA(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path, index_col=0, skiprows=3, usecols=[0, 1])
        dates_data = df.index.name.split('：')
        assert (len(dates_data) >= 2) and (dates_data[0] == '日期'), 'invalid format from TaiChuangXiShuoCTA'
        df = df.loc[['单位净值', '累计单位净值'], :].T
        df['datetime'] = pd.to_datetime(str(dates_data[1]), infer_datetime_format=True).date()
        df['fund_id'] = 'SNH765'
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        return df

    @staticmethod
    def _read_for_TaiChuangXiShuoCTANormalWithPrism(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品名称', '产品代码', '业务日期', '单位净值', '累计单位净值', '虚拟后净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        df['datetime'] = df.datetime.transform(lambda x: pd.to_datetime(str(x), infer_datetime_format=True).date())
        return df

    @staticmethod
    def _read_for_TaiChuangXiShuoCTANormal(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品名称', '产品代码', '业务日期', '单位净值', '累计单位净值', '虚拟计提净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        df['datetime'] = df.datetime.transform(lambda x: pd.to_datetime(str(x), infer_datetime_format=True).date())
        return df

    @staticmethod
    def _read_for_BoPu6(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品名称', '产品代码', '业务日期', '单位净值', '累计单位净值', '虚拟后净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        df['datetime'] = df.datetime.transform(lambda x: pd.to_datetime(str(x), infer_datetime_format=True).date())
        return df

    @staticmethod
    def _read_for_BoPu6WithoutV(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path, header=None)
        df.iloc[:, 0] = df.iloc[:, 0].map(lambda x: x.split('：')[0])
        df = df.set_index(df.columns[0])
        pat = r'\d+'
        date = re.findall(pat, df.index.array[2])
        df = df.T.loc[:, ['基金代码', '基金名称', '基金份额净值', '基金份额累计净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT).assign(datetime=pd.to_datetime('-'.join(date), infer_datetime_format=True).date())
        return df

    @staticmethod
    def _read_for_YuanHui1(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品名称', '产品代码', '业务日期', '单位净值', '累计单位净值', '虚拟后单位净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        return df

    @staticmethod
    def _read_for_ZhuoShiJiaXin(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品名称', '产品代码', '业务日期', '单位净值', '累计单位净值', '虚拟后净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        df['datetime'] = df.datetime.transform(lambda x: pd.to_datetime(str(x), infer_datetime_format=True).date())
        return df

    @staticmethod
    def _read_for_JinDeChengYiJingXin2(file_path: str) -> pd.DataFrame:
        data = []
        with pdfplumber.open(file_path) as pdf:
            first_page = pdf.pages[0]
            the_date = pd.to_datetime(re.sub('年|月|日', "-", first_page.extract_words()[5]['text'])[:-1], infer_datetime_format=True).date()
            for one in first_page.extract_tables({'vertical_strategy': 'text'})[0][1:3]:
                data.append(one[0].split())
        df = pd.Series(data[1], index=data[0]).to_frame().T
        df = df[['单位净值']]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        df['fund_id'] = 'SJJ077'
        df['datetime'] = the_date
        df['acc_unit_value'] = df['net_asset_value']
        return df

    @staticmethod
    def _read_for_ShuangLong_LongFu2(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, df.columns.intersection(['产品名称', '产品代码', '业务日期', '单位净值', '累计单位净值', '虚拟计提净值', '虚拟后净值'])]
        df = df.rename(columns=HedgeFundNAVReader.COLUMNS_DICT)
        df['datetime'] = df.datetime.transform(lambda x: pd.to_datetime(str(x), infer_datetime_format=True).date())
        return df

    def _notify_error_event(self, err_msg: str):
        print(f'[read_navs_and_dump_to_db] {err_msg}')
        self._wechat_bot.send_hedge_fund_nav_update_failed(err_msg)

    def read_navs_and_dump_to_db(self):
        try:
            with open(os.path.join(self._read_dir, UID_FILE_NAME), 'rb') as f:
                uid_last = f.read()
                if not uid_last:
                    uid_last = None
        except Exception as e:
            self._notify_error_event(f'read uid file failed (e){e}, use None instead(read all emails)')
            uid_last = None

        try:
            mar = MailAttachmentRetriever(self._read_dir)
            data = mar.get_excels(IMAP_SPType.IMAP_QQ, self._user_name, self._password, uid_last)
        except Exception as e:
            self._notify_error_event(f'FATAL ERROR!! get new data of hedge fund nav failed (e){e}')
            return

        uid_last_succeed: Optional[bytes] = None
        df_list: List[pd.DataFrame] = []
        for name, comp_date in data.items():
            uid, file_path = comp_date
            if not name.endswith('.xls') and not name.endswith('.xlsx') and not name.endswith('.pdf'):
                self._notify_error_event(f'not a valid file, do not process it (name){name} (file_path){file_path}')
                continue

            try:
                if 'SLW695' in name:
                    continue
                elif 'SGM473' in name:
                    df = HedgeFundNAVReader._read_for_HuaChengZhiYuanNo3WithoutV(file_path)
                elif '华澄致远三号' in name:
                    df = HedgeFundNAVReader._read_for_HuaChengZhiYuanNo3(file_path)
                elif 'SLC213_天演广全' in name:
                    df = HedgeFundNAVReader._read_for_TianYanGuangQuanWithoutV(file_path)
                elif '天演广全' in name:
                    df = HedgeFundNAVReader._read_for_TianYanGuangQuan(file_path)
                elif '安贤花木长盛' in name:
                    if name.endswith('.pdf'):
                        continue
                    df = HedgeFundNAVReader._read_for_AnXianHuaMuChangSheng(file_path)
                elif '吾执CTA一号' in name:
                    if name.endswith('.pdf'):
                        continue
                    df = HedgeFundNAVReader._read_for_WuZhiCTA1(file_path)
                elif '产品净值情况' in name:
                    df = HedgeFundNAVReader._read_for_WuZhiCTA1WithoutV(file_path)
                elif '信朴10号私募基金' in name:
                    df = HedgeFundNAVReader._read_for_XinPu10(file_path)
                elif 'SNH765泰创-羲朔CTA' in name:
                    df = HedgeFundNAVReader._read_for_TaiChuangXiShuoCTA(file_path)
                elif '泰创-羲朔CTA' in name:
                    if '棱镜' in name:
                        df = HedgeFundNAVReader._read_for_TaiChuangXiShuoCTANormalWithPrism(file_path)
                    else:
                        df = HedgeFundNAVReader._read_for_TaiChuangXiShuoCTANormal(file_path)
                elif '博普量化对冲6号' in name:
                    df = HedgeFundNAVReader._read_for_BoPu6(file_path)
                elif 'SEE891-发送每日净值' in name:
                    df = HedgeFundNAVReader._read_for_BoPu6WithoutV(file_path)
                elif '源晖量化一号' in name:
                    df = HedgeFundNAVReader._read_for_YuanHui1(file_path)
                elif '卓识佳鑫' in name:
                    df = HedgeFundNAVReader._read_for_ZhuoShiJiaXin(file_path)
                elif '金锝诚意精心2号' in name:
                    if '月报' in name or not name.endswith('.pdf'):
                        continue
                    df = HedgeFundNAVReader._read_for_JinDeChengYiJingXin2(file_path)
                elif '双隆-隆富2号' in name:
                    df = HedgeFundNAVReader._read_for_ShuangLong_LongFu2(file_path)
                else:
                    raise NotImplementedError('unknown hedge fund nav file from attachment')
            except Exception as e:
                self._notify_error_event(f'{e} (parse) (name){name} (file_path){file_path}')
                continue

            try:
                df = df.drop(columns=['fund_name'], errors='ignore')
                df['insert_time'] = datetime.datetime.now()
                df = self._dump_to_db(df)
                if df is not None:
                    df_list.append(df)
                else:
                    print(f'[read_navs_and_dump_to_db] duplicated data, do not process it (name){name}')
                # 走到这里都认为是已经处理完了这条数据
                uid_last_succeed = uid
                time.sleep(1)
            except Exception as e:
                traceback.print_exc()
                self._notify_error_event(f'{e} (dump) (name){name} (file_path){file_path}')
                break

        if df_list:
            try:
                whole_df = pd.concat(df_list).set_index('fund_id')
                print(whole_df)
            except Exception as e:
                self._notify_error_event(f'{e} (concat)')
                return
            else:
                # self._wechat_bot.send_hedge_fund_nav_update(whole_df)
                print(f'[read_navs_and_dump_to_db] done (uid_last){uid_last_succeed} (df){whole_df}')
        else:
            whole_df = None
            print(f'[read_navs_and_dump_to_db] no new data this time, done (uid_last){uid_last_succeed}')
        # 记录下成功的最后一个uid
        if uid_last_succeed is not None:
            with open(os.path.join(self._read_dir, UID_FILE_NAME), 'wb') as f:
                f.write(uid_last_succeed)
            return whole_df
        return

    def _dump_to_db(self, df: pd.DataFrame):
        def _check_after_merged(x: pd.Series, now_df: pd.DataFrame):
            try:
                now_data = now_df[(now_df.fund_id == x.fund_id) & (now_df.datetime == x.datetime)]
                if now_data[['net_asset_value', 'acc_unit_value', 'v_net_value']].iloc[0].astype('float64').equals(x[['net_asset_value', 'acc_unit_value', 'v_net_value']].astype('float64')):
                    return pd.Series(dtype='object')
                else:
                    return x
            except (KeyError, IndexError):
                return x

        assert df.datetime.nunique() == 1, 'should have single datetime'
        now_df = BasicDataApi().get_hedge_fund_nav(fund_id_list=list(df.fund_id.unique()))
        if now_df is not None and not now_df.empty:
            # 同产品同日期的净值如果已经存在了且没有变化，就不写DB了
            now_df = now_df.drop(columns=['update_time', 'create_time', 'is_deleted', 'change_rate']).sort_values(by=['fund_id', 'datetime']).drop_duplicates(subset=['fund_id', 'datetime'], keep='last')
            now_df = now_df.astype({'net_asset_value': 'float64', 'acc_unit_value': 'float64', 'v_net_value': 'float64'})
            df = df.reindex(columns=now_df.columns).astype(now_df.dtypes.to_dict())
            df = df.merge(now_df, how='left', on=['fund_id', 'datetime', 'net_asset_value', 'acc_unit_value', 'v_net_value'], indicator=True, validate='one_to_one')
            df = df[df._merge == 'left_only'].drop(columns=['_merge', 'insert_time_y', 'calc_date_y', 'adjusted_net_value_y', 'ta_factor_y']).rename(columns={'insert_time_x': 'insert_time', 'calc_date_x': 'calc_date', 'adjusted_net_value_x': 'adjusted_net_value', 'ta_factor_x': 'ta_factor'})
            if df.empty:
                return
            # FIXME 没想到特别好的方法 遍历每一行再check一下
            df['datetime'] = pd.to_datetime(df.datetime, infer_datetime_format=True).dt.date
            df = df.apply(_check_after_merged, axis=1, now_df=now_df)
            if df.empty:
                return
            df['insert_time'] = df.insert_time.map(lambda x: x.to_pydatetime())
            df = df.set_index(['fund_id', 'datetime'])
            print(df)
            df.update(now_df.set_index(['fund_id', 'datetime']), overwrite=False)
            df = df.reset_index()
            df['datetime'] = df.datetime.map(lambda x: x.date())
            print(df)
            # 先删后添
            for fund_id in df.fund_id.unique():
                BasicDataApi().delete_hedge_fund_nav(fund_id_to_delete=fund_id, date_list=df[df.fund_id == fund_id].datetime.to_list())
        df.to_sql(HedgeFundNAV.__table__.name, BasicDatabaseConnector().get_engine(), index=False, if_exists='append')
        return df


if __name__ == '__main__':
    try:
        email_data_dir = os.environ['SURFING_EMAIL_DATA_DIR']
        user_name = os.environ['SURFING_EMAIL_USER_NAME']
        password = os.environ['SURFING_EMAIL_PASSWORD']
    except KeyError as e:
        import sys
        sys.exit(f'can not found enough params in env (e){e}')

    hf_nav_r = HedgeFundNAVReader(email_data_dir, user_name, password)
    hf_nav_r.read_navs_and_dump_to_db()



from typing import List
import pandas as pd
import numpy as np
import json
import datetime

# 回测底层基金黑名单，筛掉没有办法处理的问题基金 2020-05-30
## #摘牌基金 发生大额申赎，导致净值剧烈波动， 用东财机构持有权重数据无法过滤
##'020035!0': 国泰上证5年期国债ETF联接A  '020036!0' 国泰上证5年期国债ETF联接C
BLACK_SHEEP = ['020035!0', '020036!0']

#基金基准发生改变  2020-06-22
##'320014!0':诺安沪深300指数增强   '050021!0':‘博时创业版ETF联接A 
BLACK_SHEEP.extend(['320014!0', '050021!0'])

#基金类别改变 2020-06-30
## ‘000512!0’: '国泰沪深300指数增强A'   曾用名 国泰结构转型
BLACK_SHEEP.extend(['000512!0'])

#基金类别改变 2020-07-14
## ‘001879!0’: '长城创业板指数增强A'   曾用名 长城创新动力灵活配置混合基金
BLACK_SHEEP.extend(['001879!0'])

#基金类别改变 2020-07-27
## ’519034!0' "海富通中证500增强" 曾用名 海富通中证低碳
BLACK_SHEEP.extend(['519034!0'])

# 发生净值突变的分级基金 属于放宽国债tag，加入信用债下的基金  2020-06-22
## 向上波动 '164210!1': 天弘同利分级, '164703!1': 汇添富互利分级 ,
## 向下波动 '000428!0' : 易方达聚盈分级
BLACK_SHEEP.extend(['164210!1','164703!1', '000428!0'])

# 没有办法筛选掉，早期终止 2020-06-30
## 000169!1 泰达宏利高票息A
## 000022!0 南方中债中期票据A
BLACK_SHEEP.extend(['000169!1','000022!0'])

# 基金类型改变 货币型 变债券型 2020-10-22 稳健策略净值波动
## 380010!0    中银理财30天a -> 中银聚享债券a
## 380011!0    中银理财30天b -> 中银聚享债券b
BLACK_SHEEP.extend(['380010!0','380011!0'])

EHR_PROBLEM_EKYWORDS = {
    'E': ['ETF', 'REITs', 'ESG', 'CES'],
    'H': ['H股', 'A-H50', '50AH', '中证AH'],
    'R': ['REITs'],
}

# 绝对收益基金关键字
ALPHA_FUND_KEY_WORD = ['对冲策略','阿尔法','绝对收益','对冲套利','量化收益']
ALPHA_PATTEN = '|'.join(ALPHA_FUND_KEY_WORD)

def filter_fund_info(fund_info: pd.DataFrame, index_list: List[str]) -> pd.DataFrame:
    # 去掉基金份额 E类 H类 R类
    # E类 指定代销机构对基金
    # H类 基金场内上市份额
    # R类 对特定投资群体进行发售，特定群体指基本养老保险基金，企业年金计划筹集的资金等
    ehr_funds = []
    # 这里itertuples会比iterrows快1个数量级
    for r in fund_info[['desc_name', 'fund_id']].itertuples():
        for type_word, problem_words in EHR_PROBLEM_EKYWORDS.items():
            i = r.desc_name
            for problem_i in problem_words:
                i = i.replace(problem_i, '')
            if type_word in i:
                ehr_funds.append(r.fund_id)

    no_mmf_index = set(index_list) - set(['mmf'])
    fund_info = fund_info[
        (
            fund_info.index_id.isin(no_mmf_index)
            # 只选非分级基金
            & ((fund_info.structure_type == 0))
            # 去掉etf基金
            & (fund_info.is_etf != 1)
            # 只选择人民币支付的基金
            & (fund_info.currency == 'CNY')
            # 只选择开放式基金，去掉封闭式基金
            & (fund_info.is_open == 1)
            # 去掉基金份额类型是 E H R
            & (~fund_info.fund_id.isin(ehr_funds))
            # 去掉超短债类型的基金，国债tag下 13只
            & (~fund_info.desc_name.str.contains('超短债'))
            # 去掉短期融资债 10只
            & (~fund_info.desc_name.str.contains('短融')) 
            # 去掉双债型债券  双债多含可转债 共41只
            & (~fund_info.desc_name.str.contains('双债'))
            # 去掉短期纯债型基金
            & (fund_info.wind_class_2 != '短期纯债型基金')
            # 货币基金总共1000多只，fund indicator 计算压力大，当前人工用2020年5月的货币基金规模取了前50只作为 货币基金池
            | (fund_info.is_selected_mmf == 1)# & (~fund_info.desc_name.str.contains('B')))TODO 回测对比
            # 增加绝对收益型基金 17只
            | (fund_info.desc_name.str.contains(ALPHA_PATTEN))
        )
        & (~fund_info.fund_id.isin(BLACK_SHEEP))
    ].copy()
    
    fund_info = fund_info[~(fund_info.wind_class_2 == '普通股票型基金')].copy()
    fund_info = fund_info[~(fund_info.wind_class_2 == '偏股混合型基金')].copy()

    return fund_info

def fund_info_update(fund_info: pd.DataFrame):
    #增加绝对收益 到货币型基金 38只
    mmf_exp_fund_list = fund_info[fund_info.desc_name.str.contains(ALPHA_PATTEN)].index.tolist()
    fund_info.loc[mmf_exp_fund_list,'index_id'] = 'mmf'
    
    #国债类扩容
    fund_info.loc[fund_info.national_debt_extension == 1, 'index_id'] = 'national_debt'
            
    #增加港股基金 基准1含有恒生 共69只
    hsi_fund = []
    for r in fund_info.itertuples():
        if (r.benchmark_1) and ('恒生' in r.benchmark_1):
            hsi_fund.append(r.fund_id)
    fund_info.loc[fund_info.fund_id.isin(hsi_fund), 'index_id_new'] = 'hsi'

    #增加混合债型一级基金 到国债下，共271 （混合一级含股票， 股灾时mdd会扩大，还是考虑排除）
    #hybrid_class_1_funds = fund_info[fund_info['wind_class_2'] == '混合债券型一级基金'].fund_id.to_list()
    #fund_info.loc[fund_info.fund_id.isin(hybrid_class_1_funds), 'index_id_new'] = 'national_debt'
    fund_info.loc[:,'index_id'] = fund_info['index_id_new']

    return fund_info

def active_fund_info(fund_info: pd.DataFrame, fund_benchmark: pd.DataFrame):
    #active_stock_fund_list = ['普通股票型基金', '偏股混合型基金', '增强指数型基金','灵活配置型基金', '平衡混合型基金']
    active_stock_fund_list = ['普通股票型基金', '偏股混合型基金'] # 只保留普通股票型基金 和 偏股股票型基金
    stock_index = ['csi500','hs300']
    gem_index_list = ['cnext_lowv_bc','cnext_mome_gr','cnext_r','gei','gem','gem_50'] #创业板相关基准
    
    # 过滤基金 参考fund_info
    # 去掉基金份额 E类 H类 R类
    # E类 指定代销机构对基金
    # H类 基金场内上市份额
    # R类 对特定投资群体进行发售，特定群体指基本养老保险基金，企业年金计划筹集的资金等
    ehr_funds = []
    # 这里itertuples会比iterrows快1个数量级
    for r in fund_info[['desc_name', 'fund_id']].itertuples():
        for type_word, problem_words in EHR_PROBLEM_EKYWORDS.items():
            i = r.desc_name
            for problem_i in problem_words:
                i = i.replace(problem_i, '')
            if type_word in i:
                ehr_funds.append(r.fund_id)
    
    fund_info = fund_info[
        (
            # 只选非分级基金
            ((fund_info.structure_type == 0))
            # 去掉etf基金
            & (fund_info.is_etf != 1)
            # 只选择人民币支付的基金
            & (fund_info.currency == 'CNY')
            # 只选择开放式基金，去掉封闭式基金
            & (fund_info.is_open == 1)
            # 去掉基金份额类型是 E H R
            & (~fund_info.fund_id.isin(ehr_funds))
        )]
    fund_info = fund_info.set_index('fund_id').drop(['benchmark'], axis=1)
    fund_benchmark = fund_benchmark.set_index('fund_id')
    fund_info = fund_info.join(fund_benchmark)[['desc_name','start_date','end_date','wind_class_1','wind_class_2','benchmark_s_raw']].copy()
    stock_fund = fund_info[fund_info['wind_class_2'].isin(active_stock_fund_list)]
    res = []
    for r in stock_fund.itertuples():
        i = r.benchmark_s_raw
        if isinstance(i, str):
            i = json.loads(i)
            _n = 1
            dic = {'fund_id':r.Index}
            for index_id, w in i.items():
                dic[f'fund_{_n}'] = index_id
                dic[f'weight_{_n}'] = w
                _n += 1
            res.append(dic)
        elif pd.isnull(i):
            res.append({})     
    _benchmark_info = pd.DataFrame(res).dropna(subset=['fund_id']).set_index('fund_id')
    stock_fund = stock_fund.join(_benchmark_info).copy()
    stock_fund = stock_fund.dropna(subset=['weight_1']).copy()
    res = []
    for r in stock_fund.itertuples():
        tag_wait_list = []
        if r.weight_1 >= 0.5:
            tag_wait_list.append(r.fund_1)
        if r.weight_2 >= 0.5:
            tag_wait_list.append(r.fund_2)
        tag_list = [_i for _i in tag_wait_list if _i in stock_index]
        if len(tag_list) > 0:
            tag = tag_list[0]
        else:
            gem_list = [_i for _i in tag_wait_list if _i in gem_index_list]
            if len(gem_list) > 0:
                tag = gem_list[0]
            else:
                tag = None
        res.append(tag)
    stock_fund['index_id'] = res
    _res = []
    for index_id in ['hs300','csi500','gem']:
        _df = stock_fund[(stock_fund['index_id'] == index_id)]
        _res.append(_df)
    active_fund_info = pd.concat(_res)[['desc_name','start_date','end_date','wind_class_1','wind_class_2','index_id','weight_1']].rename(columns={'weight_1':'weight'}).reset_index()
    
    ###
    #人工增加中证500 hs300 主动型， 基金基准人工扩容 TODO
    csi500 = ['000001!0',
                '000021!0',
                '004065!0',
                '004272!0',
                '004273!0',
                '005188!0',
                '005189!0',
                '005355!0',
                '005356!0',
                '005385!0',
                '040007!0',
                '070002!0',
                '110011!0',
                '257010!0',
                '550009!0']
    
    hs300 = ['000011!0', '000551!0', '000761!0', '090003!0']
    fund_info.loc[csi500, 'index_id'] = 'csi500'
    fund_info.loc[hs300, 'index_id'] = 'hs300' 
    csi500_df = fund_info.loc[csi500][['desc_name','start_date','end_date','wind_class_1','wind_class_2','index_id']].reset_index()
    hs300_df = fund_info.loc[hs300][['desc_name','start_date','end_date','wind_class_1','wind_class_2','index_id']].reset_index()
    active_fund_info = active_fund_info.append(csi500_df).append(hs300_df)
    ###

    # 只选择成立一年的
    _begin_date = datetime.datetime.today().date() - datetime.timedelta(days=365)
    active_fund_info = active_fund_info[active_fund_info['start_date'] < _begin_date]
    return active_fund_info

def get_conv_funds(fund_conv_stats, fund_info):
    conv_df = fund_conv_stats.pivot_table(index='datetime', values='conv_weights', columns='fund_id')
    conv_funds = conv_df.columns[conv_df.max() > 20].tolist()
    conv_funds = fund_info[(fund_info['fund_id'].isin(conv_funds)) 
              & (fund_info.structure_type == 0)].fund_id.tolist()
    return conv_funds